"""Here different kinds of constraint functions are defined.

"""

import logging
import logging.config
from abc import ABC, abstractmethod
from os import path
from typing import Callable, List

import numpy as np

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


class ConstraintError(Exception):
    """Raised when an error related to the Constraint class in encountered.

    """


class ConstraintBase(ABC):
    """Base class for constraints.

    """

    @abstractmethod
    def evaluate(
        self, decision_vector: np.ndarray, objective_vector: np.ndarray
    ) -> float:
        """Evaluate the constraint functions and return a float
        indicating how severely the constraint has been broken.

        Args:
            decision_vector (np.ndarray): A decision_vector containing
            the decision variable values.
            objective_vector (np.ndarray): A decision_vector containing the
            objective function values.

        Returns:
            float: A float representing how and if the constraing has
            been violated. A positive value represents no violation and a
            negative value represents a violation. The absolute value of the
            returned float functions as an indicator of the severity of the
            violation (or how well the constraint holds, if the returned value
            of positive).

d        """
        pass


class ScalarConstraint(ConstraintBase):
    """A simple scalar constraint that evaluates to a single scalar.

    Args:
        name (str): Name of the constraint.
        n_decision_vars (int): Number of decision variables present in the
        constraint.
        n_objective_funs (int): Number of objective functions present in
        the constraint.
        evaluator (Callable): A callable to evaluate the constraint.

    Attributes:
        name (str): Name of the constraint.
        n_decision_vars (int): Number of decision variables present in the
        constraint.
        n_objective_funs (int): Number of objective functions present in
        the constraint.
        evaluator (Callable): A callable to evaluate the constraint.

    """

    def __init__(
        self,
        name: str,
        n_decision_vars: int,
        n_objective_funs: int,
        evaluator: Callable,
    ) -> None:
        self.__name: str = name
        self.__n_decision_vars: int = n_decision_vars
        self.__n_objective_funs: int = n_objective_funs
        self.__evaluator: Callable = evaluator

    @property
    def name(self) -> str:
        return self.__name

    @property
    def n_decision_vars(self) -> int:
        return self.__n_decision_vars

    @property
    def n_objective_funs(self) -> int:
        return self.__n_objective_funs

    @property
    def evaluator(self) -> Callable:
        return self.__evaluator

    def evaluate(
        self, decision_vector: np.ndarray, objective_vector: np.ndarray
    ) -> float:
        """Evaluate the constraint and return a float indicating how and if the
        constraint was violated. A negative value indicates a violation and
        a positive value indicates a non-violation.

        Args:
            decision_vector (np.ndarray): A decision_vector containing the
            values of the decision variables.
            objective_vector (np.ndarray): A decision_vector containing the
            values of the objective functions.

        Returns:
            float: A float indicating how the constraint holds.

        Raises:
            ConstraintError: When something goes wrong evaluating the
            constraint or the objectives and decision vectors are of wrong
            shape.

        """
        decision_l = (
            len(decision_vector)
            if decision_vector.ndim == 1
            else decision_vector.shape[1]
        )
        if decision_l != self.__n_decision_vars:
            msg = (
                "Decision decision_vector {} is of wrong lenght: "
                "Should be {}, but is {}"
            ).format(decision_vector, self.__n_decision_vars, decision_l)
            logger.error(msg)
            raise ConstraintError(msg)

        objective_l = (
            len(objective_vector)
            if objective_vector.ndim == 1
            else objective_vector.shape[1]
        )
        if objective_l != self.__n_objective_funs:
            msg = (
                "Objective decision_vector {} is of wrong lenght:"
                " Should be {}, but is {}"
            ).format(objective_vector, self.__n_objective_funs, objective_l)
            logger.error(msg)
            raise ConstraintError(msg)
        try:
            result = self.__evaluator(decision_vector, objective_vector)
        except (TypeError, IndexError) as e:
            msg = (
                "Bad arguments {} and {} supplied to the evaluator:" " {}"
            ).format(str(decision_vector), objective_vector, str(e))
            logger.error(msg)
            raise ConstraintError(msg)

        return result


supported_operators: List[str] = ["==", "<", ">"]
"""List[str]: Shows the operators supportted in the
constraint_function_factory."""


def constraint_function_factory(
    lhs: Callable, rhs: float, operator: str
) -> Callable:
    """A function that creates an evaluator to be used with the ScalarConstraint
    class. Constraints should be formulated in a way where all the mathematical
    expression are on the left hand side, and the constants on the right hand
    side.

    Args:
       lhs (Callable): The left hand side of the constraint. Should be a
       callable function representing a mathematical expression.
       rhs (float): The right hand side of a constraint. Represents the right
       hand side of the constraint.
       operator (str): The kind of constraint. Can be '==', '<', '>'.

    Returns:
       Callable: A function that can be called to evaluate the rhs and
       which returns representing how the constraint is obeyed. A negative
       value represent a violation of the constraint and a positive value an
       agreement with the constraint. The absolute value of the float is a
       direct indicator how the constraint is violated/agdreed with.

    Raises:
        ValueError: The supplied operator is not supported.

    """
    if operator not in supported_operators:
        msg = "The operator {} supplied is not supported.".format(operator)
        logger.error(msg)
        raise ValueError(msg)

    if operator == "==":

        def equals(
            decision_vector: np.ndarray, objective_vector: np.ndarray
        ) -> float:
            return -abs(lhs(decision_vector, objective_vector) - rhs)

        return equals

    elif operator == "<":

        def lt(
            decision_vector: np.ndarray, objective_vector: np.ndarray
        ) -> float:
            return rhs - lhs(decision_vector, objective_vector)

        return lt

    elif operator == ">":

        def gt(
            decision_vector: np.ndarray, objective_vector: np.ndarray
        ) -> float:
            return lhs(decision_vector, objective_vector) - rhs

        return gt

    else:
        # if for some reason a bad operator falls through
        msg = "Bad operator argument supplied: {}".format(operator)
        logger.error(msg)
        raise ValueError(msg)



"""Define numerical methods to be used in the solvers.
"""
import logging
import logging.config
from abc import ABC, abstractmethod
from os import path
from typing import Any, Callable, Dict, Optional, Tuple, Union

import numpy as np
from scipy.optimize import differential_evolution

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


class NumericalMethodError(Exception):
    """Raised when an error is encountered in the numerical method classes.

    """


class NumericalMethodBase(ABC):
    """Define an abstract class for all methods to follow

    Args:
        method(Callable[[Any], np.ndarray]): A callable method that minimizes a
            given (set of) funtions and returns the ideal solution(s).

    Attributes:
        method(Callable[[Any], np.ndarray]): A callable method that minimizes a
        given (set of) funtions and returns the ideal solution(s).

    """

    def __init__(self, method: Callable[[Any], np.ndarray]):
        self.__method = method

    @property
    def method(self) -> Any:
        return self.__method

    @method.setter
    def method(self, val: Any):
        self.__method = val

    @abstractmethod
    def run(
        self,
        evaluator: Callable[..., np.ndarray],
        bounds: np.ndarray,
        evaluator_args: Union[Dict[str, Any], Tuple[Any]] = {},
        variables: Optional[np.ndarray] = None,
        objectives: Optional[np.ndarray] = None,
    ):
        """Passes the appropiate parameters to the underlying method and
        returns the solution given by the method.

        Args:
            evaluator(Callable[[np.ndarray, Any], np.ndarray]):
            A function to be evaluated and minimized.
            bounds(np.ndarray): The bounds of the variables as a 2D array with
            each row representing the lower (first column) and upper (last
            column) bounds of each variable.
            evaluator_args(Optional[Dict[str, Any]]): An optional tuple
            containing positional arguments to be passed to the evaluator.

        Returns:
            The solution computed by the numerical method.

        """
        pass


class ScipyDE(NumericalMethodBase):
    """Uses the differential solver implemented in SciPy.

    Attributes:
        method_kwargs(Optional[Any]): The keyword arguments to be passed to the
        differential evolution algorithm implemented in SciPy, see the SciPy
        documentation for more info.

    Args:
        method_kwargs(Optional[Any]): The keyword arguments to be passed to the
        differential evolution algorithm implemented in SciPy, see the SciPy
        documentation for more info.

    """

    def __init__(
        self,
        method_kwargs: Optional[Any] = {
            "tol": 0.000001,
            "popsize": 10,
            "maxiter": 50000,
            "polish": True,
        },
    ):
        super().__init__(differential_evolution)
        self.method_kwargs = method_kwargs

    @property
    def method_kwargs(self) -> Any:
        return self.__method_kwargs

    @method_kwargs.setter
    def method_kwargs(self, val: Any):
        self.__method_kwargs = val

    def run(
        self,
        evaluator: Callable[[np.ndarray, Any], np.ndarray],
        bounds: np.ndarray,
        evaluator_args: Union[Dict[str, Any], Tuple[Any]] = {},
        variables: Optional[np.ndarray] = None,
        objectives: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Run the differential solver minimizinf the given evaluator and
        following the given variable bounds.

        Args:
            evaluator(Callable[[np.ndarray, Any], np.ndarray]):
            A function to be evaluated and minimized.
            bounds(np.ndarray): The bounds of the variables as a 2D array with
            each row representing the lower (first column) and upper (last
            column) bounds of each variable.
            evaluator_args(Union[Dict[str, Any], Tuple[Any], None]): An
            optional tuple containing positional arguments to be passed to the
            evaluator.

        Returns:
            np.ndarray: An array containing the optimal solution reached by the
            differential evolution method.

        Raises:
            NumericalMethodError: Something goes wrong with the evaluator or
            the minimization.

        """
        if evaluator_args is None:
            results = self.method(evaluator, bounds, **self.method_kwargs)
        else:
            results = self.method(
                evaluator, bounds, args=evaluator_args, **self.method_kwargs
            )

        if results.success:
            return results.x
        else:
            msg = (
                "The differential solver was not successful. " "Reason: {}"
            ).format(results.message)
            logger.debug(msg)
            raise NumericalMethodError(msg)


class DiscreteMinimizer(NumericalMethodBase):
    """Finds the minimum value from a discrete set of pre-defined points.

    """

    def __init__(self):
        super().__init__(self.minimizer)

    def minimizer(
        self,
        evaluator: Callable[..., np.ndarray],
        bounds: np.ndarray,
        variables: Optional[np.ndarray] = None,
        objectives: Optional[np.ndarray] = None,
        kwargs: Dict[str, Any] = {},
    ) -> np.ndarray:
        """
        Args:
            evaluator (Callable[[np.ndarray, Any], np.ndarray]):
            A function to be evaluated and minimized.
            bounds (np.ndarray): The bounds of the variables as a 2D array with
            each row representing the lower (first column) and upper (last
            column) bounds of each variable.
            variables (Optional[np.ndarray]): 2D array with the variables of
            the discrete problem.
            objectives (Optional[np.ndarray]): 2D array with the objective
            function values of the discrete problem.
            kwargs (Optional[Dict[str, Any]]): Extra arguments to be passed to
            the evaluator.

        Returns:
            np.ndarray: The variables that results in the objective values that
            minimize the evaluator.

        Raises:
            NumericalMethodError: variables or objectives not specified.

        """
        if variables is None:
            msg = "Variables must be specified for the minimizer to work."
            logger.error(msg)
            raise NumericalMethodError(msg)
        if objectives is None:
            msg = "Objectives must be specified for the minimizer to work."
            logger.error(msg)
            raise NumericalMethodError(msg)

        if bounds is not None:
            mask_lower_bounds = np.all(
                np.greater(variables, bounds[:, 0]), axis=1
            )
            mask_upper_bounds = np.all(
                np.less(variables, bounds[:, 1]), axis=1
            )
            mask_feasible = np.logical_and(
                mask_lower_bounds, mask_upper_bounds
            )

            feasible_objectives = objectives[mask_feasible]
            feasible_variables = variables[mask_feasible]
        else:
            feasible_objectives = objectives
            feasible_variables = variables

        res = evaluator(feasible_variables, feasible_objectives, **kwargs)
        idx = np.argmin(res)

        return feasible_variables[idx]

    def run(
        self,
        evaluator: Callable[[np.ndarray, Any], np.ndarray],
        bounds: np.ndarray,
        evaluator_args: Union[Dict[str, Any], Tuple[Any]] = {},
        variables: Optional[np.ndarray] = None,
        objectives: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Run the minimizer minimizing the given evaluator and
        following the given variable bounds.

        Args:
            evaluator(Callable[[np.ndarray, Any], np.ndarray]):
            A function to be evaluated and minimized.
            bounds(np.ndarray): The bounds of the variables as a 2D array with
            each row representing the lower (first column) and upper (last
            column) bounds of each variable.
            evaluator_args(Union[Dict[str, Any], Tuple[Any], None]): An
            optional tuple containing positional arguments to be passed to the
            evaluator.

        Returns:
            np.ndarray: An array containing the optimal solution reached by the
            differential evolution method.

        Raises:
            NumericalMethodError: Something goes wrong with the evaluator or
            the minimization.

        """
        results = self.method(
            evaluator,
            bounds,
            variables=variables,
            objectives=objectives,
            kwargs=evaluator_args,
        )

        return results



"""Define different solvers that solve a scalarized version of a MOO problem.

"""
import abc
import logging
import logging.config
from abc import abstractmethod
from os import path
from typing import Any, Optional, Tuple, Union

import numpy as np

from desdeov2.problem.Problem import ProblemBase
from desdeov2.solver.ASF import ASFBase
from desdeov2.solver.NumericalMethods import (
    DiscreteMinimizer,
    NumericalMethodBase,
    ScipyDE,
)

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


class ScalarSolverError(Exception):
    """Raised when an error related to the solver classes is
    encountered.

    """

    pass


class ScalarSolverBase(abc.ABC):
    """A base class to define the interface for the various solvers that solve
    a scalarized version of the original MOO problem.

    Args:
        problem (ProblemBase): The underlaying problem object with the
        specifications of the problem to solve.
        method (NumericalMethodBase): The numerical method to solve the
        scalarizing functions.

    Attributes:
        problem (ProblemBase): The underlaying problem object with the
        specifications of the problem to solve.
        method (NumericalMethodBase): The numerical method to solve the
        scalarizing functions.

    """

    def __init__(
        self, problem: Union[ProblemBase], method: NumericalMethodBase
    ):
        self.__problem = problem
        self.__method = method

    @property
    def problem(self) -> ProblemBase:
        return self.__problem

    @problem.setter
    def problem(self, val: ProblemBase):
        self.__problem = val

    @property
    def method(self) -> NumericalMethodBase:
        return self.__method

    @method.setter
    def method(self, val: NumericalMethodBase):
        self.__method = val

    @abstractmethod
    def _evaluator(
        self,
        decision_vectors: np.ndarray,
        objective_vectors: Optional[np.ndarray],
    ) -> np.ndarray:
        """A helper function to formulate the scalarized version of the
        underlying MOO problem

        """
        pass

    @abstractmethod
    def solve(
        self, args: Any
    ) -> Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Solves the problem and returns relevant information.

        Args:
            *args(Any): Any kind of arguments that are relevant to the solver.
            See the solvers for further details.

        Returns:
            Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]: A tuple
            containing the decision variables as the first element and the
            evaluation result of the underlaying porblem as the second element.

        """
        pass


class WeightingMethodScalarSolver(ScalarSolverBase):
    """A class to represent a solver for solving a problems using the
    weighting method.

    Args:
        problem (ProblemBase): The underlaying problem object with the
        specifications of the problem to solve.
        method (NumericalMethodBase): The numerical method to solve the
        scalarizing functions.

    Attributes:
        weights (np.ndarray): The weights corresponsing to each objectie in the
        problem.

    """

    def __init__(self, problem: ProblemBase, method: NumericalMethodBase):
        super().__init__(problem, method)
        self.__weights: np.ndarray = None

    @property
    def weights(self) -> np.ndarray:
        return self.__weights

    @weights.setter
    def weights(self, val: np.ndarray):
        self.__weights = val

    def _evaluator(
        self,
        decision_vectors: np.ndarray,
        objective_vectors: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """A helper function to transform the output of the MOO problem
        to a weighted sum.

        Args:
            decision_vectors (np.ndarray): An array of arrays representing
            decision variable values to evaluate the MOO problem.

        Returns:
            np.ndarray: An array with the weighted sums corresponding to
            each solution evaluated using the decision variable vectors
            present in decision_vectors.

        Note:
            If any of the constraints are broken for a particular solution, the
            corresponding weighted sum is set to infinity, regardless of the
            weight value. It is therefore assumed that each objective is to be
            minimized.

        """
        constraint_vectors: np.ndarray
        if objective_vectors is None:
            objective_vectors, constraint_vectors = self.problem.evaluate(
                decision_vectors
            )
        else:
            constraint_vectors = None

        weighted_sums = np.zeros(len(objective_vectors))
        for ind, elem in enumerate(objective_vectors):
            if constraint_vectors is not None and np.any(
                constraint_vectors[ind] < 0
            ):
                # If any of the constraints is broken, set the sum to
                # infinity
                weighted_sums[ind] = np.inf
            else:
                # Compute the weigted sum for each objective vector using the
                # dot product
                weighted_sums[ind] = np.dot(
                    objective_vectors[ind], self.__weights
                )

        return weighted_sums

    def solve(
        self, weights: np.ndarray
    ) -> Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Use differential evolution to solve the weighted sum problem.

        Args:
            weights (np.ndarray): Array of weights to weigh each objective in
            the sum.

        Returns:
            Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]: A tuple
            containing the decision variables as the first element and the
            evaluation results of the underlaying porblem as the second
            element.

        Note:
            This method might invoke runtime warnings, which are most likely
            related to the usage of infinity (np.inf).

        """
        self.weights = weights
        bounds = self.problem.get_variable_bounds()

        if isinstance(self.method, DiscreteMinimizer):
            x = self.method.run(
                self._evaluator,
                bounds,
                variables=self.problem.decision_vectors,
                objectives=self.problem.objective_vectors,
            )
        elif isinstance(self.method, ScipyDE):
            # Scipy does not work with vectors, so we just feed it the first
            # value in the vector returned by the evaluator
            func = lambda x, *args: self._evaluator(x)[0]  # noqa: E731
            x = self.method.run(func, bounds)

        return (x, self.problem.evaluate(x))


class EpsilonConstraintScalarSolver(ScalarSolverBase):
    """A class to represent a solver for solving porblems using the epsilon
    constraint method.

    Attributes:
        epsilons (np.ndarray): The epsilon values to set as the upper limit
        for each objective when treated as a constraint.
        to_be_minimized (int): Integer representing which objective function
        should be minimized.

    """

    def __init__(self, problem: ProblemBase, method: NumericalMethodBase):
        super().__init__(problem, method)
        self.__epsilons: np.ndarray
        self.__to_be_minimized: int = 0

    @property
    def epsilons(self) -> np.ndarray:
        return self.__epsilons

    @epsilons.setter
    def epsilons(self, val: np.ndarray):
        """Set the epsilon values

        Args:
            epsilons (np.ndarray): Array of the epsilon values.

        Raises:
            ScalarSolverError: The array with the epsilon values is of the
            wrong length.

        """
        if len(val) != self.problem.n_of_objectives:
            msg = (
                "The length of the epsilons array '{}' must match the "
                "number of objectives '{}'."
            ).format(len(val), self.problem.n_of_objectives)
            logger.debug(msg)
            raise ScalarSolverError(msg)

        self.__epsilons = val

    @property
    def to_be_minimized(self) -> int:
        return self.__to_be_minimized

    @to_be_minimized.setter
    def to_be_minimized(self, val: int):
        self.__to_be_minimized = val

    def _evaluator(
        self,
        decision_vector: np.ndarray,
        objective_vectors: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """A helper function to express the original problem as an epsilon
           constraint problem.

        Args:
            decision_vectors (np.ndarray): An array of arrays representing the
            decision variable values to evaluate the underlaying MOO problem.

        Returns:
            np.ndarray: An array of the epsilon constraint problem values
            corresponding to each decision variable vector.

        Note:
            The epsilons attribute  must be set before this function works.

        """

        constraints: np.ndarray
        epsilon_values: np.ndarray

        if objective_vectors is None:
            objective_vectors, constraints = self.problem.evaluate(
                decision_vector
            )
        else:
            constraints = None

        epsilon_values = np.zeros(len(objective_vectors))

        for (ind, elem) in enumerate(objective_vectors):
            if constraints is not None and np.any(constraints[ind] < 0):
                # suicide method for broken constraints
                epsilon_values[ind] = np.inf
                continue
            else:
                mask = np.array(
                    [i != self.to_be_minimized for i in range(len(elem))]
                )
                if np.any(elem[mask] > self.epsilons[mask]):
                    epsilon_values[ind] = np.inf
                else:
                    epsilon_values[ind] = elem[self.to_be_minimized]

        return epsilon_values

    def solve(self, to_be_minimized: int):
        """Use differential evolution to solve the epsilon constraint problem.

        Args:
            to_be_minimized (int): The index of the problem in the underlaying
            MOProblem to be solved.

        Returns:
            Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]: A tuple
            containing the decision variables as the first element and the
            evaluation results of the underlaying problem as the second
            element.

        """
        self.to_be_minimized = to_be_minimized
        bounds = self.problem.get_variable_bounds()

        if isinstance(self.method, DiscreteMinimizer):
            x = self.method.run(
                self._evaluator,
                bounds,
                variables=self.problem.decision_vectors,
                objectives=self.problem.objective_vectors,
            )
        elif isinstance(self.method, ScipyDE):
            # Scipy does not work with vectors, so we just feed it the first
            # value in the vector returned by the evaluator
            func = lambda x, *args: self._evaluator(x)[0]  # noqa: E731
            x = self.method.run(func, bounds)

        return (x, self.problem.evaluate(x))


class ASFScalarSolver(ScalarSolverBase):
    """A class to represent a solver tha uses the achievement scalarizing
    method to solve a multiobjective optimization problem.

    Attributes:
        asf (ASFBase): The ASF to be used in the scalarization of the
        underlying MOO problem.
        reference_point (np.ndarray): The reference point (or similar) that is
        used in the ASF.

    """

    def __init__(self, problem: ProblemBase, method: NumericalMethodBase):
        super().__init__(problem, method)
        self.__asf: ASFBase
        self.__reference_point: np.ndarray

    @property
    def asf(self) -> ASFBase:
        return self.__asf

    @asf.setter
    def asf(self, val: ASFBase):
        self.__asf = val

    @property
    def reference_point(self) -> np.ndarray:
        return self.__reference_point

    @reference_point.setter
    def reference_point(self, val: np.ndarray):
        self.__reference_point = val

    def _evaluator(
        self,
        decision_vectors: np.ndarray,
        objective_vectors: Optional[np.ndarray] = None,
        constraints: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """A helper function to express the original problem as an ASF
           problem.

        Args:
            decision_vectors (np.ndarray): An array of arrays representing the
            decision variable values to evaluate the underlaying MOO problem.
            objective_vectors (Optional[np.ndarray]): An array of arrays
            representing the objective vector values of the problem.
            constraints (Optional[np.ndarray]): An array of arrays containing
            the constraints values evaluated with each entry of decision
            vectors and objective vectors

        Returns:
            np.ndarray: An array of the ASF problem values corresponding to
            each decision variable vector.

        Note:
            Requires an achievement scalarizing function to be defined and set
            to the ASFScalarSolver object. See ASF.py for available functions
            and further details.

        """

        asf_values: np.ndarray
        if objective_vectors is None and constraints is None:
            objective_vectors, constraints = self.problem.evaluate(
                decision_vectors
            )
        elif constraints is not None:
            constraints = constraints

        asf_values = np.zeros(len(objective_vectors))  # type: ignore

        for (ind, elem) in enumerate(objective_vectors):  # type: ignore
            if constraints is not None and np.any(constraints[ind] < 0):
                # suicide method for broken constraints
                asf_values[ind] = np.inf
            else:
                asf_values[ind] = self.asf(elem, self.reference_point)

        if np.all(asf_values == np.inf):
            logger.warning(
                "All asf values are inf, result may " "be not sensical."
            )

        return asf_values

    def solve(
        self, reference_point: np.ndarray
    ) -> Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Use differential evolution to solve the ASF problem.

        Args:
            reference_point (np.ndarray): An array representing the reference
            point in the objective function space used in the is ASF.
        Returns:
            Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]: A tuple
            containing the decision variables as the first element and the
            evaluation results of the underlaying problem as the second
            element.

        """
        self.reference_point = reference_point
        bounds = self.problem.get_variable_bounds()

        if isinstance(self.method, DiscreteMinimizer):
            x = self.method.run(
                self._evaluator,
                bounds,
                evaluator_args={
                    "constraints": self.problem.evaluate_constraint_values()
                },
                variables=self.problem.decision_vectors,
                objectives=self.problem.objective_vectors,
            )
        elif isinstance(self.method, ScipyDE):
            # Scipy does not work with vectors, so we just feed it the first
            # value in the vector returned by the evaluator
            func = lambda x, *args: self._evaluator(x)[0]  # noqa: E731
            x = self.method.run(func, bounds)

        return (x, self.problem.evaluate(x))



"""Define a decorator to prevent the setting of new attributes for objects
outside of the object's __init__ function. The ability to define these new
attributes outside of __init__ is a very common source for bugs. Therefore, we
get rid of it with the cost a bit of extra overhead, but the payoff in the long
run should be definitely worth it.

Thanks to Yoann's response on the StackOverflow thread I found. My version if a
slight modification with an extra argument given to the decorator for logging
purposes.

(Just complete the SO link)
Source: /questions/3603502/prevent-creating-new-attributes-outside-init

"""

from functools import wraps
from logging import Logger
from typing import Any


def frozen(logger: Logger) -> Any:
    """A decorator to prevent the definition of new attributes outside of that
    object's __init__ function.

    Attributes:
        cls (Class): The class to be decorated
        logger (Logger): A logger object to log debug messages

    Returns:
        Class: A class with frozen attributes

    Usage:
        @frozen(loggername)
        class foo(parent):
            ...

    Note: Typehinting this thing seems to be a nightmare.

    """

    def _frozen(cls: Any) -> Any:
        cls.__frozen = False

        def frozensetattr(self, key, value) -> None:
            """Prevent the definition of new attributes. Log a message if a non
            existing attribute is attempted to be defined.

            """
            if self.__frozen and not hasattr(self, key):
                msg = "Class {} is frozen. Cannot set {} = {}".format(
                    cls.__name__, key, value
                )
                logger.debug(msg)
            else:
                object.__setattr__(self, key, value)

        def init_decorator(func: Any) -> Any:
            @wraps(func)
            def wrapper(self, *args, **kwargs):
                func(self, *args, **kwargs)
                self.__frozen = True

            return wrapper

        cls.__setattr__ = frozensetattr
        cls.__init__ = init_decorator(cls.__init__)

        return cls

    return _frozen



"""Various interactive multi-objective optimization methods belonging to the
NAUTILUS-family are defined here.

"""

import logging
import logging.config
from os import path
from typing import List, Optional, Tuple, Union

import numpy as np
from sklearn.cluster import KMeans
from sklearn.metrics import pairwise_distances_argmin_min

from desdeov2.methods.InteractiveMethod import (
    InteractiveMethodBase,
    InteractiveMethodError,
)
from desdeov2.problem.Problem import (
    ProblemBase,
    ScalarDataProblem,
    ScalarMOProblem,
)
from desdeov2.solver.ASF import ReferencePointASF
from desdeov2.solver.NumericalMethods import NumericalMethodBase, ScipyDE
from desdeov2.solver.PointSolver import IdealAndNadirPointSolver
from desdeov2.solver.ScalarSolver import (
    ASFScalarSolver,
    EpsilonConstraintScalarSolver,
)
from desdeov2.utils.frozen import frozen

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


@frozen(logger)
class Nautilus(InteractiveMethodBase):
    """Implements the basic NAUTILUS methods as presented in `Miettinen 2010`_

    Attributes:
        epsilon (float): A small number used in calculating the utopian point.
        itn (int): Total number of iterations.
        h (int): Current iteration.
        ith (int): Number of remaining iterations.
        lower_bound (List[np.ndarray]): The lower bounds of reachable values
        from different iteration point.
        upper_bound (List[np.ndarray]): The upper bounds of reachable values
        from different iteration point.
        zs (List[np.ndarray]): Iteration points.
        q (np.ndarray): Current iteration point.
        xs (List[np.ndarray]): The solutions for each iteration.
        fs (List[np.ndarray]): The objective vector values for each iteration.
        ds (List[float]): The distance to the pareto front in each iteration.
        preference_index_set (np.ndarray): The current DM's preference as index
        sets
        preference_percentages (np.ndarray): The current DM's preferences as
        percentages.
        mu (np.ndarray): The current preferential factors calculated from the
        DM preferences.
        scalar_solver (ASFScalarSolver): A solver for solving scalarized
        problems using ASFs.
        asf (ReferencePointASF): Reference point based ASF used in the
        scalar_solver

    .. _Miettinen 2010:
        Miettinen, K.; Eskelinen, P.; Ruiz, F. & Luque, M.
        NAUTILUS method: An interactive technique in multiobjective
        optimization based on the nadir point
        Europen Joural of Operational Research, 2010, 206, 426-434

    """

    def __init__(self, problem: ProblemBase):
        super().__init__(problem)

        # Used to calculate the utopian point from the ideal point
        self.__epsilon: float = 0.0
        self.__itn: int = 0  # total number of iterations
        self.__h: int = 0  # current iteration
        self.__ith: int = 0  # number of remaining iterations
        self.__lower_bounds: List[np.ndarray] = []
        self.__upper_bounds: List[np.ndarray] = []
        self.__zs: List[np.ndarray] = []  # iteration points
        self.__q: np.ndarray = None  # The current reference point
        self.__xs: List[np.ndarray] = []  # solutions for each iteration
        self.__fs: List[np.ndarray] = []  # objectives for each iteration
        self.__ds: List[float] = []  # distances for each iteration
        self.__preference_index_set: np.ndarray = None
        self.__preference_percentages: np.ndarray = None
        self.__mu: np.ndarray = np.zeros(  # current preferential factors
            self.problem.n_of_objectives
        )

        self.__numerical_method: NumericalMethodBase = ScipyDE()
        self.__scalar_solver: ASFScalarSolver = ASFScalarSolver(
            self.problem, self.__numerical_method
        )  # noqa
        self.__asf: ReferencePointASF = ReferencePointASF(None, None, None)
        self.__scalar_solver.asf = self.__asf

        self.__epsilon_solver: EpsilonConstraintScalarSolver = EpsilonConstraintScalarSolver(  # noqa: E501
            self.problem, self.__numerical_method
        )

        # flags for the iteration phase
        self.__use_previous_preference: bool = False
        self.__step_back: bool = False
        self.__short_step: bool = False
        self.__first_iteration: bool = True

    @property
    def epsilon(self) -> float:
        return self.__epsilon

    @epsilon.setter
    def epsilon(self, val: float):
        self.__epsilon = val

    @property
    def itn(self) -> int:
        return self.__itn

    @itn.setter
    def itn(self, val: int):
        """Set the number of total iterations to be carried.

        Args:
            val (int): The total number of iterations. Must be positive.

        Raises:
            InteractiveMethodError: val is not positive.

        """
        if val < 0:
            msg = (
                "The number of total iterations "
                "should be positive. Given iterations '{}'".format(str(val))
            )
            logger.debug(msg)
            raise InteractiveMethodError(msg)
        self.__itn = val

    @property
    def h(self) -> int:
        return self.__h

    @h.setter
    def h(self, val: int):
        self.__h = val

    @property
    def ith(self) -> int:
        return self.__ith

    @ith.setter
    def ith(self, val: int):
        """Set the number of remaining iterations. Should be less than the current
        remaining iterations

        Args:
            val (int): New number of iterations to carry out.

        Raises:
            InteractiveMethodError: val is either negative or greater than the
            current number of remaining iterations.

        """
        if val < 0:
            msg = (
                "The given number of iterations left "
                "should be positive. Given iterations '{}'".format(str(val))
            )
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        elif val > self.__ith:
            msg = (
                "The given number of iterations '{}' left should be less "
                "than the current number of iterations left '{}'"
            ).format(val, self.__ith)
            logger.debug(msg)
            raise InteractiveMethodError(msg)
        self.__ith = val

    @property
    def lower_bounds(self) -> List[np.ndarray]:
        return self.__lower_bounds

    @lower_bounds.setter
    def lower_bounds(self, val: List[np.ndarray]):
        self.__lower_bounds = val

    @property
    def upper_bounds(self) -> List[np.ndarray]:
        return self.__upper_bounds

    @upper_bounds.setter
    def upper_bounds(self, val: List[np.ndarray]):
        self.__upper_bounds = val

    @property
    def zs(self) -> List[np.ndarray]:
        return self.__zs

    @zs.setter
    def zs(self, val: List[np.ndarray]):
        self.__zs = val

    @property
    def q(self) -> np.ndarray:
        return self.__q

    @q.setter
    def q(self, val: np.ndarray):
        self.__q = val

    @property
    def xs(self) -> List[np.ndarray]:
        return self.__xs

    @xs.setter
    def xs(self, val: List[np.ndarray]):
        self.__xs = val

    @property
    def fs(self) -> List[np.ndarray]:
        return self.__fs

    @fs.setter
    def fs(self, val: List[np.ndarray]):
        self.__fs = val

    @property
    def ds(self) -> List[float]:
        return self.__ds

    @ds.setter
    def ds(self, val: List[float]):
        self.__ds = val

    @property
    def mu(self) -> np.ndarray:
        return self.__mu

    @mu.setter
    def mu(self, val: np.ndarray):
        self.__mu = val

    @property
    def preference_index_set(self) -> np.ndarray:
        return self.__preference_index_set

    @preference_index_set.setter
    def preference_index_set(self, val: np.ndarray):
        """Set the indexes to rank each of the objectives in order of
        importance.

        Args:
            val (np.ndarray): A 1D-vector containing the preference index
            corresponding to each obejctive.

        Raise:
            InteractiveMethodError: The length of the index vector does not
            match the number of objectives in the problem.

        """
        if len(val) != self.problem.n_of_objectives:
            msg = (
                "The number of indices '{}' does not match the number "
                "of objectives '{}'"
            ).format(len(val), self.problem.n_of_objectives)
            logger.debug(msg)
            raise InteractiveMethodError(msg)
        elif not (1 <= max(val) <= self.problem.n_of_objectives):
            msg = (
                "The minimum index of importance must be greater or equal "
                "to 1 and the maximum index of improtance must be less "
                "than or equal to the number of objectives in the "
                "problem, which is {}. Check the indices {}"
            ).format(self.problem.n_of_objectives, val)
            logger.debug(msg)
            raise InteractiveMethodError(msg)
        self.__preference_index_set = val

    @property
    def preference_percentages(self) -> np.ndarray:
        return self.__preference_percentages

    @preference_percentages.setter
    def preference_percentages(self, val: np.ndarray):
        """Set the percentages to descripe how to improve each objective.

        Args:
            val (np.ndarray): A 1D-vector containing percentages corresponding
            to each objective.

        Raises:
            InteractiveMethod: The lenght of the prcentages vector does not
            match the number of objectives in the problem.

        """
        if len(val) != self.problem.n_of_objectives:
            msg = (
                "The number of given percentages '{}' does not match the "
                "number of objectives '{}'"
            ).format(len(val), self.problem.n_of_objectives)
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        elif np.sum(val) != 100:
            msg = (
                "The sum of the percentages must be 100. Current sum" " is {}."
            ).format(np.sum(val))
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        self.__preference_percentages = val

    @property
    def scalar_solver(self) -> ASFScalarSolver:
        return self.__scalar_solver

    @scalar_solver.setter
    def scalar_solver(self, val: ASFScalarSolver):
        self.__scalar_solver = val

    @property
    def asf(self) -> ReferencePointASF:
        return self.__asf

    @asf.setter
    def asf(self, val: ReferencePointASF):
        self.__asf = val

    def _calculate_iteration_point(self) -> np.ndarray:
        """Calculate the new iteration point based on the current iteration's
        iteration point and pareto optimal solution.

        Returns:
            np.ndarray: New iteration point

        """
        # Number of iterations left. The plus one is to ensure the current
        # iteration is also counted.
        ith = self.ith
        z_prev = self.zs[self.h]  # == from previous step
        f_h = self.fs[self.h + 1]  # current step

        z_h = ((ith - 1) / (ith)) * z_prev + (1 / (ith)) * f_h

        return z_h

    def _calculate_distance(self) -> float:
        """Calculate the distance to the pareto optimal front in the current
        iteration.

        Returns:
            float: The distance to the pareto front ranging from 0-100, 100
            being the closest distance to the front.

        """
        ds = 100 * (
            np.linalg.norm(self.zs[self.h + 1] - self.problem.nadir)
            / (np.linalg.norm(self.fs[self.h + 1] - self.problem.nadir))
        )
        return ds

    def initialize(  # type: ignore
        self, itn: int = 5
    ) -> Tuple[np.ndarray, List[Tuple[float, float]], float]:
        """Initialize the method by setting the initialization parameters and
        caluclating the initial bounds of the problem, the nadir and ideal
        point, if not defined in the problem.

        Args:
            itn (int): Number of total iterations. Defaults to 5.

        Returns:
            Tuple[np.ndarray, List[Tuple[float, float]], float]: A tuple
            containing:
                np.ndarray: The current iteration point.
                List[Tuple[float, float]]: A list with tuples with the lower
                and upper bounds for the next iteration
                float: The distance of the current iteration point to the
                pareto optimal set.

        """
        self.itn = itn

        # Check if the ideal and nadir points are set
        if self.problem.ideal is None and self.problem.nadir is None:
            # Both missing, compute both
            solver = IdealAndNadirPointSolver(
                self.problem, self.__numerical_method
            )
            self.problem.ideal, self.problem.nadir = solver.solve()

        elif self.problem.ideal is None:
            # ideal missing, compute it
            solver = IdealAndNadirPointSolver(
                self.problem, self.__numerical_method
            )
            self.problem.ideal, _ = solver.solve()

        elif self.problem.nadir is None:
            # nadir missing, compute it
            solver = IdealAndNadirPointSolver(
                self.problem, self.__numerical_method
            )
            _, self.problem.nadir = solver.solve()

        self.h = 0
        # Skip the checks in the setter of ith
        self.__ith = self.itn

        self.zs = [None] * (self.itn + 1)
        self.zs[0] = self.problem.nadir

        self.lower_bounds = [None] * (self.itn + 1)
        self.lower_bounds[self.h] = self.problem.ideal

        self.upper_bounds = [None] * (self.itn + 1)
        self.upper_bounds[self.h] = self.problem.nadir

        self.xs = [None] * (self.itn + 1)

        self.fs = [None] * (self.itn + 1)

        self.ds = [0.0] * (self.itn + 1)

        self.asf.nadir = self.problem.nadir
        self.asf.utopian_point = self.problem.ideal - self.epsilon

        return (
            self.zs[self.h],
            list(zip(self.lower_bounds[self.h], self.upper_bounds[self.h])),
            self.ds[self.h],
        )

    def iterate(self) -> Tuple[np.ndarray, List[Tuple[float, float]], float]:
        """Iterate once according to the user preference given in the
        interaction phase.

        Returns:
            Tuple[np.ndarray, List[Tuple[float, float]], float]: A tuple
            containing:
                np.ndarray: The current iteration point.
                List[Tuple[float, float]]: A list with tuples with the lower
                and upper bounds for the next iteration
                float: The distance of the current iteration point to the
                pareto optimal set.

        Raises:
            InteractiveMethodError: If the preferential factors can't be
            computed

        Note:
            The current iteration is to be interpreted as self.h + 1, since
            incrementation of the current iteration happens in the interaction
            phase.

            If both the relative importance and percentages are defined,
            percentages are used.

        """
        # Calculate the preferential factors or use existing ones
        if not self.__short_step:
            if self.preference_percentages is not None:
                # use percentages to calcualte the new iteration point
                delta_q = self.preference_percentages / 100
                self.mu = 1 / (
                    delta_q
                    * (
                        self.problem.nadir
                        - (self.problem.ideal - self.epsilon)
                    )
                )

            elif self.preference_index_set is not None:
                # Use the relative importance to calcualte the new points
                print(self.preference_index_set)
                for (i, r) in enumerate(self.preference_index_set):
                    self.mu[i] = 1 / (
                        r
                        * (
                            self.problem.nadir[i]
                            - (self.problem.ideal[i] - self.epsilon)
                        )
                    )

            elif self.__use_previous_preference:
                # previous
                pass

            else:
                msg = "Could not compute the preferential factors."
                logger.debug(msg)
                raise InteractiveMethodError(msg)

        if not self.__short_step and not self.__use_previous_preference:
            # Take a normal step and calculate a new reference point
            # set the current iteration point as the reference point
            self.q = self.zs[self.h]

            # set the preferential factors in the underlaying asf
            self.asf.preferential_factors = self.mu

            # solve the ASF
            (solution, (objective, _)) = self.scalar_solver.solve(self.q)

            # Store the solution and corresponding objective vector
            self.xs[self.h + 1] = solution
            self.fs[self.h + 1] = objective[0]

            # calculate a new iteration point
            self.zs[self.h + 1] = self._calculate_iteration_point()

        elif not self.__step_back and self.__use_previous_preference:
            # Use the solution and objective of the last step
            self.xs[self.h + 1] = self.xs[self.h]
            self.fs[self.h + 1] = self.fs[self.h]
            self.zs[self.h + 1] = self._calculate_iteration_point()

        else:
            # Take a short step
            # Update the current iteration point
            self.zs[self.h + 1] = (
                0.5 * self.zs[self.h + 1] + 0.5 * self.zs[self.h]
            )

        # calculate the lower bounds for the next iteration
        if self.ith > 1:
            # last iteration, no need to calculate these
            self.lower_bounds[self.h + 1] = np.zeros(
                self.problem.n_of_objectives
            )

            self.__epsilon_solver.epsilons = self.zs[self.h + 1]

            for r in range(self.problem.n_of_objectives):
                (_, (objective, _)) = self.__epsilon_solver.solve(r)
                self.lower_bounds[self.h + 1][r] = objective[0][r]

                # set the upper bounds
                self.upper_bounds[self.h + 1] = self.zs[self.h]

        else:
            self.lower_bounds[self.h + 1] = [
                None
            ] * self.problem.n_of_objectives
            self.upper_bounds[self.h + 1] = [
                None
            ] * self.problem.n_of_objectives

        # Calculate the distance to the pareto optimal set
        self.ds[self.h + 1] = self._calculate_distance()

        return (
            self.zs[self.h + 1],
            list(
                zip(
                    self.lower_bounds[self.h + 1],
                    self.upper_bounds[self.h + 1],
                )
            ),
            self.ds[self.h + 1],
        )

    def interact(
        self,
        index_set: np.ndarray = None,
        percentages: np.ndarray = None,
        use_previous_preference: bool = False,
        new_remaining_iterations: Optional[int] = None,
        step_back: bool = False,
        short_step: bool = False,
    ) -> Union[int, Tuple[np.ndarray, np.ndarray]]:
        """Handle user preference and set appropiate flags for the next iteration.

        Args:
            index_set (np.ndarray): An array with integers describing the
            relative importance of each objective. The integers vary between 1
            and the maximum number of objectives in the problem.
            percentages (np.ndarray): Percentages describing the absolute
            importance of each objective. The sum of these must equal 100.
            use_previous_preference (bool): Use the preference
            infromation. Cannot be true during the first iteration.
            defined in the last iteration. Defaults to false.
            new_remaining_iterations (int): Set a new number of remaining
            iterations to be carried. Must be positive and not exceed the
            current number of iterations left.
            step_back (bool): Step from the previous point in the next
            iteration. Cannot step back from first iteration.
            short_step (bool): When step_back, take a shorter step in the same
            direction as in the previous iteration from the previous
            iteration's iteration point. Can only short step when stepping
            back.

        Returns:
            Union[int, Tuple[np.ndarray, np.ndarray]]: The number of remaining
            iteration. If this function is envoked after the last iteration,
            returns a tuple with the optimal solution and objective values.

        Raises:
            InteractiveMethodError: Some of the arguments are not set
            correctly. See the documentation for the arguments.

        """
        if index_set is not None:
            self.preference_index_set = index_set
            self.__use_previous_preference = False

        elif percentages is not None:
            self.preference_percentages = percentages
            self.__use_previous_preference = False

        elif self.mu is not None and use_previous_preference:
            self.__use_previous_preference = use_previous_preference

        elif step_back and short_step:
            # no preference needed for short step
            pass

        else:
            msg = "Cannot figure out preference infromation."
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        if new_remaining_iterations is not None:
            self.ith = new_remaining_iterations

        if not step_back:
            self.__step_back = False
            # Advance the current iteration, if not the first iteration
            if not self.__first_iteration:
                self.ith -= 1
                self.h += 1
            else:
                self.__first_iteration = False

            if self.ith == 0:
                # Last iteration, terminate the solution
                return (self.xs[self.h], self.fs[self.h])

        else:
            if self.__first_iteration:
                msg = "Cannot take a backwards step on the first iteration."
                logger.debug(msg)
                raise InteractiveMethodError(msg)
            self.__step_back = True

        if short_step:
            if not step_back:
                msg = (
                    "Can take a short step only when stepping from the "
                    "previous point."
                )
                logger.debug(msg)
                raise InteractiveMethodError(msg)
        self.__short_step = short_step

        return self.ith


@frozen(logger)
class ENautilus(InteractiveMethodBase):
    """Implements the enhanced Nautilus variant, E-Nautilus originally
    presented in `Ruiz 2015`_

    Args:
        pareto_front (np.ndarray): The representation of a pareto front for a
        problem to be solved.
        objective_vectors (np.ndarray): The objective vectors corresponding to
        the pareto front.
        nadir (np.ndarray): The nadir point.
        ideal (np.ndarray): The ideal point.
        n_iters (int): The number of total iterations.
        n_points (int): The number of intermediate points to be generated.
        zshi (np.ndarray): The intermediate points at each iteration.
        h (int): The current iteration.
        ith (int): The iterations left.
        par_sub (List[np.ndarray]): The subspace of the pareto front reachable
        during each iteration.
        par_obj (List[np.ndarray]): The subspace of the reachable objective
        vector values during each iteration.
        fhilo (np.ndarray): The lower bounds of the reachable set from each
        intermediate point.
        d (np.ndarray): How close to the pareto front the each of the
        intermediate points are for each iteration.
        zpref (np.ndarray): The preferred point at each iteration.

    .. _Ruiz 2015:
        Ruiz A. B.; Sindhya K.; Miettinen K.; Ruiz F. & Luque M.
        E-NAUTILUS: A decision support system for complex multiobjective
        optimization problems based on the NAUTILUS method
        Europen Joural of Operational Research, 2015, 246, 218-231
    """

    def __init__(self, problem: ProblemBase):
        super().__init__(problem)

        if isinstance(problem, ScalarMOProblem):
            msg = (
                "Currently E-NAUTILUS works only with the "
                "ScalarDataProblem problem class or by sypplying "
                "the pareto set and objectives manually."
            )
            logger.error(msg)
            raise NotImplementedError(msg)

        # The full pareto front
        self.__pareto_front: np.ndarray = None
        # Corresponding pareto optimal objective vectors for the full front
        self.__objective_vectors: np.ndarray = None
        self.__nadir: np.ndarray = None
        self.__ideal: np.ndarray = None
        # number of total iterations
        self.__n_iters: int = 0
        # number of points to be shown to the DM after each iteration
        self.__n_points: int = 0
        # the i (n_points) intermediate points at step h
        self.__zshi: np.ndarray = None
        # current iteration
        self.__h: int = 0
        # iterations left
        self.__ith: int = 0
        # the subset of the reachable pareto optimal solutions
        # during each iteration
        self.__par_sub: List[np.ndarray] = []
        # the subset of the reachable pareto optimal objective vectors
        # during each iteration
        self.__obj_sub: List[np.ndarray] = []
        # lower bounds of the reachable set from each intermediate point
        self.__fhilo: np.ndarray = None
        # closeness to the pareto front
        self.__d: np.ndarray = None
        # prefered point
        self.__zpref: np.ndarray = None

    @property
    def pareto_front(self) -> np.ndarray:
        return self.__pareto_front

    @pareto_front.setter
    def pareto_front(self, val: np.ndarray):
        self.__pareto_front = val

    @property
    def objective_vectors(self) -> np.ndarray:
        return self.__objective_vectors

    @objective_vectors.setter
    def objective_vectors(self, val: np.ndarray):
        self.__objective_vectors = val

    @property
    def nadir(self) -> np.ndarray:
        return self.__nadir

    @nadir.setter
    def nadir(self, val: np.ndarray):
        """Set the nadir point.

        Args:
            val (np.ndarray): The nadir point.

        Raises:
            InteractiveMethodError: The nadir point is of the wrong dimensions.

        """
        if len(val) != self.objective_vectors.shape[1]:
            msg = (
                "The nadir point's length '{}' must match the number of "
                "objectives '{}' (columns) specified in the given "
                "pareto front."
            ).format(len(val), self.objective_vectors.shape[1])
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        self.__nadir = val

    @property
    def ideal(self) -> np.ndarray:
        return self.__ideal

    @ideal.setter
    def ideal(self, val: np.ndarray):
        """Set the ideal point.

        Args:
            val (np.ndarray): The ideal point.

        Raises:
            InteractiveMethodError: The ideal point is of the wrong dimensions.

        """
        if len(val) != self.objective_vectors.shape[1]:
            msg = (
                "The ideal point's length '{}' must match the number of "
                "objectives '{}' (columns) specified in the given "
                "pareto front."
            ).format(len(val), self.objective_vectors.shape[1])
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        self.__ideal = val

    @property
    def n_iters(self) -> int:
        return self.__n_iters

    @n_iters.setter
    def n_iters(self, val: int):
        """Set the total number of iterations to be carried out.

        Args:
            val (int): The number of iterations.

        Raises:
            InteractiveMethodError: The number of iterations is non-positive.

        """
        if val < 1:
            msg = "Number of iterations must be greater than zero."
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        self.__n_iters = val

    @property
    def n_points(self) -> int:
        return self.__n_points

    @n_points.setter
    def n_points(self, val: int):
        """The number of points to be presented to the DM during each iteration.

        Args:
            val (int): Number of points to be shown.

        Raises:
            InteractiveMethodError: The number of points is non-positive.

        """
        if val < 1:
            msg = "The number of points shown must be greater than zero."
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        self.__n_points = val

    @property
    def zshi(self) -> np.ndarray:
        return self.__zshi

    @zshi.setter
    def zshi(self, val: np.ndarray):
        self.__zshi = val

    @property
    def h(self) -> int:
        return self.__h

    @h.setter
    def h(self, val: int):
        self.__h = val

    @property
    def ith(self) -> int:
        return self.__ith

    @ith.setter
    def ith(self, val: int):
        self.__ith = val

    @property
    def par_sub(self) -> List[np.ndarray]:
        return self.__par_sub

    @par_sub.setter
    def par_sub(self, val: List[np.ndarray]):
        self.__par_sub = val

    @property
    def obj_sub(self) -> List[np.ndarray]:
        return self.__obj_sub

    @obj_sub.setter
    def obj_sub(self, val: List[np.ndarray]):
        self.__obj_sub = val

    @property
    def fhilo(self) -> np.ndarray:
        return self.__fhilo

    @fhilo.setter
    def fhilo(self, val: np.ndarray):
        self.__fhilo = val

    @property
    def d(self) -> np.ndarray:
        return self.__d

    @d.setter
    def d(self, val: np.ndarray):
        self.__d = val

    @property
    def zpref(self) -> np.ndarray:
        return self.__zpref

    @zpref.setter
    def zpref(self, val: np.ndarray):
        self.__zpref = val

    def initialize(  # type: ignore
        self,
        n_iters: int,
        n_points: int,
        pareto_front: Optional[np.ndarray] = None,
        objective_vectors: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Initialize the method with the input data required by E-NAUTILUS by
        either the given problem in the initializer of the class or explicitly
        given pareto and objective data.

        Args:
            n_iters (int): The number of total iterations to be carried out.
            n_points (int): The number of points to be shown to the DM each
            iteration.
            pareto_front (Optional[np.ndarray]): Vectors preresenting the
            pareto optimal solutions of a MOO problem.
            objective_vectors (Optional[np.ndarray]): The objective vectors
            that correspond to the pareto optimal solutions.

        Returns:
            Tuple[np.ndarray, np.ndarray]: A tuple containing:
                np.ndarray: The nadir point of the problem.
                np.ndarray: The ideal point of the problem.
        Raises:
            NotImplementedError: The problem type must be ScalarDataProblem, if
            not, raise this.

        Note:
            Only support solving ScalarDataProblem at the moment.

        """
        if isinstance(self.problem, ScalarDataProblem):
            self.pareto_front = self.problem.decision_vectors
            self.objective_vectors = self.problem.objective_vectors
        else:
            msg = (
                "ENautilus currently supports only "
                "solving ScalarDataProblems"
            )
            logger.error(msg)
            raise NotImplementedError(msg)

        self.n_iters = n_iters
        self.n_points = n_points

        # find the nadir and ideal points
        self.nadir = np.max(self.objective_vectors, axis=0)
        self.ideal = np.min(self.objective_vectors, axis=0)

        # initialize the intermediate points, bounds, and distances
        self.zshi = np.full(
            (self.n_iters, self.n_points, self.objective_vectors.shape[1]),
            np.nan,
            dtype=np.float,
        )

        self.fhilo = np.full(
            (self.n_iters, self.n_points, self.objective_vectors.shape[1]),
            np.nan,
            dtype=np.float,
        )

        self.d = np.full((self.n_iters, self.n_points), np.nan, dtype=np.float)

        # initialize the reachable subsets (None at the zeroth position,
        # because we want to use h to access them)
        self.par_sub = [None] * (self.n_iters)
        self.obj_sub = [None] * (self.n_iters)

        self.h = 0
        self.ith = self.n_iters
        self.par_sub[self.h] = self.pareto_front
        self.obj_sub[self.h] = self.objective_vectors
        self.zpref = self.nadir

        return self.nadir, self.ideal

    def iterate(self) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate the intermediate points and the lower bounds of
        the reachable solutions from each point in the next iteration
        according to a preference point specified by the decision maker during
        the previous interaction. During the first iteration, the nadir point
        is used instead.

        Returns:
            Tuple[np.ndarray, np.ndarray]: A tuple containing:
                np.ndarray: The intermediate points.
                np.ndarray: The lower bounds of the reachable values from each
                intermediate point.

        Note:
            If the requested number of points cannot be generated, the
            resulting arrays containing the intermediate points and bounds will
            be padded with NaNs.

        """
        if self.ith < 1:
            logger.info(
                (
                    "Last iteration, or trying to iterate past the last "
                    "iteration. Please call "
                    "interact with the final preference to generate the final "
                    "solution. Returning the most recent "
                    "intermediate points."
                )
            )
            return self.zshi[self.h - 1], self.fhilo[self.h - 1]

        # Use clustering to find the most representative points in the
        # objective subspace
        if self.n_points <= len(self.obj_sub[self.h]):
            kmeans = KMeans(n_clusters=self.n_points)
            kmeans.fit(self.obj_sub[self.h])
            # find the closest points to the centroids
            closest, _ = pairwise_distances_argmin_min(kmeans.cluster_centers_,
                                                       self.obj_sub[self.h])
            zbars = self.obj_sub[self.h][closest]
        else:
            # the subspace has less or an equal amount of points to the number
            # points to be shown, just use the subspace
            msg = (
                "Could not generate the requested amount of intermediate "
                "points '{}'. Generating only '{}' points"
            ).format(self.n_points, len(self.obj_sub[self.h]))
            logger.info(msg)
            zbars = self.obj_sub[self.h]

        # calculate the intermediate points
        self.zshi[self.h][0 : len(zbars)] = (  # noqa
            (self.ith - 1) / self.ith
        ) * self.zpref + (1 / self.ith) * zbars

        # calculate the lower bounds
        for r in range(self.objective_vectors.shape[1]):
            col_mask = np.full(self.objective_vectors.shape[1], True)
            col_mask[r] = False
            for i in range(len(zbars)):
                mask = np.all(
                    self.zshi[self.h, i, col_mask]
                    >= self.obj_sub[self.h][:, col_mask],
                    axis=1,
                )

                # if the mask if full of false, do nothing
                if not np.all(~mask):
                    self.fhilo[self.h, i, r] = np.min(
                        self.obj_sub[self.h][mask, r]
                    )

        # calculate the distances to the pareto front for each representative
        # point
        self.d[self.h][0 : len(zbars)] = (  # noqa
            np.linalg.norm(
                self.zshi[self.h][0 : len(zbars)] - self.nadir, axis=1  # noqa
            )
            / np.linalg.norm(zbars - self.nadir, axis=1)
        ) * 100

        return self.zshi[self.h], self.fhilo[self.h]

    def interact(  # type: ignore
        self, preferred_point: np.ndarray, lower_bounds: np.ndarray
    ) -> Union[int, Tuple[np.ndarray, np.ndarray]]:
        """Specify the next preferred point from which to iterate in the next
        iteration. The lower bounds of the reachable values from the preferred
        point are also expected. This point does not necessarely need to be a
        point returned by the iterate method in this class.

        Args:
            preferred_point (np.ndarray): An objective value vector
            representing the preferred point.
            lower_bounds (np.ndarray): The lower bounds of the reachable values
            from the preferred point.

        Returns:
            Union[int, Tuple[np.ndarray, np.ndarray]]: The number of iterations
                left, if not invoked on the last iteration. Otherwise a tuple
                containing: np.ndarray: The final pareto optimal solution.
                np.ndarray: The corresponding objevtive vector to the pareto
                optimal solution.

        Raises:
            InteractiveMethodError: The dimensions of either the preferred
            point or the lower bounds of the reachable values are incorrect.

        """
        if len(preferred_point) != self.objective_vectors.shape[1]:
            # check that the dimensions of the given points are correct
            msg = (
                "The dimensions of the prefered point '{}' do not match "
                "the shape of the objective vectors '{}'."
            ).format(len(preferred_point), self.objective_vectors.shape[1])
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        if len(lower_bounds) != self.objective_vectors.shape[1]:
            msg = (
                "The dimensions of the lower bounds for the prefered "
                "point '{}' do not match "
                "the shape of the objective vectors '{}'."
            ).format(len(lower_bounds), self.objective_vectors.shape[1])
            logger.debug(msg)
            raise InteractiveMethodError(msg)

        self.zpref = preferred_point

        if self.ith <= 1:
            # stop the algorithm and return the final solution and the
            # corresponding objective vector
            idx = np.linalg.norm(
                self.obj_sub[self.h] - self.zpref, axis=1
            ).argmin()
            self.ith = 0
            return self.par_sub[self.h][idx], self.obj_sub[self.h][idx]

        # Calculate the new reachable pareto solutions and objective vectors
        # from zpref
        cond1 = np.all(
            np.less_equal(lower_bounds, self.obj_sub[self.h]), axis=1
        )
        cond2 = np.all(np.less_equal(self.obj_sub[self.h], self.zpref), axis=1)

        indices = (cond1 & cond2).nonzero()

        self.obj_sub[self.h + 1] = self.obj_sub[self.h][indices]
        self.par_sub[self.h + 1] = self.par_sub[self.h][indices]

        self.ith -= 1
        self.h += 1

        return self.ith


class ENautilusB(ENautilus):
    """A version of E-NAUTILUS which will only calculate the reachable bounds
    for a given point.

    """
    def iterate(self) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate the distance to the pareto front and the bounds of the
        reachable values from the given point.

        Returns:
            Tuple[np.ndarray, np.ndarray]: A tuple containing:
                np.ndarray: The distance to the pareto front
                np.ndarray: The lower bounds of the reachable values from the
                intermediate point.

        """
        # calculate the lower bounds, just use the 0th index
        for r in range(self.objective_vectors.shape[1]):
            col_mask = np.full(self.objective_vectors.shape[1], True)
            col_mask[r] = False
            for i in range(len(self.zpref)):
                mask = np.all(
                    self.zpref[col_mask]
                    >= self.obj_sub[0][:, col_mask],
                    axis=1,
                )

                # if the mask if full of false, do nothing
                if not np.all(~mask):
                    self.fhilo[0, i, r] = np.min(
                        self.obj_sub[0][mask, r]
                    )

        # calculate the distances to the pareto front for each representative
        # point
        closest, _ = pairwise_distances_argmin_min(self.zpref.reshape(1, -1),
                                                   self.obj_sub[0])
        zbar = self.obj_sub[0][closest]

        self.d[0][0] = (  # noqa
            np.linalg.norm(
                self.zpref - self.nadir  # noqa
            ) / np.linalg.norm(
                zbar - self.nadir
            )
        ) * 100

        return self.fhilo[0][0], self.d[0][0]

    def interact(  # type: ignore
        self, preferred_point: np.ndarray
    ) -> None:
        """Specify the next preferred point from which to iterate in the next
        iteration.

        Args:
            preferred_point (np.ndarray): An objective value vector
            representing the preferred point.

        """
        self.zpref = preferred_point



"""Defines Objective classes to be used in Problems

"""

import logging
import logging.config
from abc import ABC, abstractmethod
from os import path
from typing import Callable

import numpy as np

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


class ObjectiveError(Exception):
    """Raised when an error related to the Objective class is encountered.

    """


class ObjectiveBase(ABC):
    """The abstract base class for objectives.

    """

    @abstractmethod
    def evaluate(self, decision_vector: np.ndarray) -> float:
        """Evaluates the objective according to a decision variable vector.

        Args:
            variables (np.ndarray): A vector of Variables to be used in
            the evaluation of the objective.

        """
        pass


class ScalarObjective(ObjectiveBase):
    """A simple objective function that returns a scalar.

    Args:
        name (str): Name of the objective.
        evaluator (Callable): The function to evaluate the objective's value.
        lower_bound (float): The lower bound of the objective.
        upper_bound (float): The upper bound of the objective.

    Attributes:
        name (str): Name of the objective.
        value (float): The current value of the objective function.
        evaluator (Callable): The function to evaluate the objective's value.
        lower_bound (float): The lower bound of the objective.
        upper_bound (float): The upper bound of the objective.

    Raises:
        ObjectiveError: When ill formed bounds are given.

    """

    def __init__(
        self,
        name: str,
        evaluator: Callable,
        lower_bound: float = -np.inf,
        upper_bound: float = np.inf,
    ) -> None:
        # Check that the bounds make sense
        if not (lower_bound < upper_bound):
            msg = (
                "Lower bound {} should be less than the upper bound " "{}."
            ).format(lower_bound, upper_bound)
            logger.error(msg)
            raise ObjectiveError(msg)

        self.__name: str = name
        self.__evaluator: Callable = evaluator
        self.__value: float = 0.0
        self.__lower_bound: float = lower_bound
        self.__upper_bound: float = upper_bound

    @property
    def name(self) -> str:
        return self.__name

    @property
    def value(self) -> float:
        return self.__value

    @value.setter
    def value(self, value: float):
        self.__value = value

    @property
    def evaluator(self) -> Callable:
        return self.__evaluator

    @property
    def lower_bound(self) -> float:
        return self.__lower_bound

    @property
    def upper_bound(self) -> float:
        return self.__upper_bound

    def evaluate(self, decision_vector: np.ndarray) -> float:
        """Evaluate the objective functions value.

        Args:
            variables (np.ndarray): A vector of variables to evaluate the
            objective function with.
        Returns:
            float: The evaluated value of the objective function.

        Raises:
            ObjectiveError: When a bad argument is supplies to the evaluator.

        """
        try:
            result = self.__evaluator(decision_vector)
        except (TypeError, IndexError) as e:
            msg = "Bad argument {} supplied to the evaluator: {}".format(
                str(decision_vector), str(e)
            )
            logger.error(msg)
            raise ObjectiveError(msg)

        # Store the value of the objective
        self.value = result

        return result



"""Contains classes representing variables to be used with the Problem
classes. This class may be used for demonstrative problems. It is not a
necessary class to be used.

"""

import logging
import logging.config
from os import path
from typing import Tuple

import numpy as np

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


class VariableError(Exception):
    """Raised when an error is encountered during the handling of the
    Variable objects.

    """

    pass


class Variable:
    """Simple variable with a name, initial value and bounds.

    Args:
        name (str): Name of the variable
        initial_value (float): The initial value of the variable.
        lower_bound (float, optional): Lower bound of the variable. Defaults
            to negative infinity.
        upper_bound (float, optional): Upper bound of the variable. Defaults
            to positive infinity.

    Attributes:
        name (str): Name of the variable.
        initial_value (float): Initial value of the variable.
        lower_bound (float): Lower bound of the variable.
        upper_bound (float): Upper bound of the variable.
        current_value (float): The current value the variable holds.

    Raises:
        VariableError: Bounds are incorrect.

    """

    def __init__(
        self,
        name: str,
        initial_value: float,
        lower_bound: float = -np.inf,
        upper_bound: float = np.inf,
    ) -> None:

        self.__name: str = name
        self.__initial_value: float
        self.__lower_bound: float
        self.__upper_bound: float
        self.__current_value: float  # NOTE: This is probably a useless attr
        # Check that the bounds make sense
        if not (lower_bound < upper_bound):
            msg = (
                "Lower bound {} should be less than the upper bound " "{}."
            ).format(lower_bound, upper_bound)
            logger.error(msg)
            raise VariableError(msg)

        # Check that the initial value is between the bounds
        if not (lower_bound < initial_value < upper_bound):
            msg = (
                "The initial value {} should be between the "
                "upper ({}) and lower ({}) bounds.".format(
                    initial_value, lower_bound, upper_bound
                )
            )
            logger.error(msg)
            raise VariableError(msg)

        self.__lower_bound = lower_bound
        self.__upper_bound = upper_bound
        self.__initial_value = initial_value
        self.__current_value = initial_value

    @property
    def name(self) -> str:
        return self.__name

    @property
    def initial_value(self) -> float:
        return self.__initial_value

    @property
    def current_value(self) -> float:
        return self.__current_value

    @current_value.setter
    def current_value(self, value: float):
        self.__current_value = value

    def get_bounds(self) -> Tuple[float, float]:
        """Return the bounds of the variables as a tuple.

        Returns:
            tuple(float, float): A tuple consisting of (lower_bound,
            upper_bound)

        """
        return (self.__lower_bound, self.__upper_bound)



"""Implementation of solvers to find certain set of points.

"""
import abc
import logging
import logging.config
from abc import abstractmethod
from os import path
from typing import Tuple

import numpy as np

from desdeov2.problem.Problem import ProblemBase
from desdeov2.solver.NumericalMethods import NumericalMethodBase

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


class PointSolverError(Exception):
    """Raised when an error related to the PointSolvers is encountered.

    """

    pass


class PointSolverBase(abc.ABC):
    """A base class that solve for particular points (or group of points) that
    solve a MOO problem, or satisfy other conditions.

    """

    @abstractmethod
    def solve(self) -> np.ndarray:
        """Solves for the points and returns them in an array."""
        pass


class IdealAndNadirPointSolver(PointSolverBase):
    """Calculates the ideal and utopian points for a given problem. The ideal
    point is calculated by minimizing each function separately.

    Args:
        problem (ProblemBase): The underlaying problem object with the
        specifications of the problem to solve.

    Attributes:
        problem (ProblemBase): The underlaying problem object with the
        specifications of the problem to solve.

    """

    def __init__(self, problem: ProblemBase, method: NumericalMethodBase):
        self.__problem = problem
        self.__method = method

    @property
    def problem(self) -> ProblemBase:
        return self.__problem

    @problem.setter
    def problem(self, val: ProblemBase):
        self.__problem = val

    @property
    def method(self) -> NumericalMethodBase:
        return self.__method

    @method.setter
    def method(self, val: NumericalMethodBase):
        self.__method = val

    def _evaluator(
        self, decision_vector: np.ndarray, objective_index: int
    ) -> float:
        """A helper function to evaluate the problem and return the value of
        the specified objective.

        Arguments:
            decision_vector (np.ndarray): The decision variables to evaluate
            the problem with.
            objective_index (int): The index of the objective whose value is to
            returned.

        Returns:
            float: The value of the specified objective function.

        """
        objective_vals, constraint_vals = self.problem.evaluate(
            decision_vector
        )
        if constraint_vals is not None and np.any(constraint_vals < 0):
            return np.inf
        else:
            return objective_vals[:, objective_index]

    def solve(self) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate the ideal and estimate of the nadir point for the MOO
        problem using a pay-off table.

        Returns:
            np.ndarray: An array with two elements. The first element is the
            ideal point specified by the diagonal of the generated pay-off
            table. The second element is an estimate of the nadir specified by
            taking the maximum value of each of the columns in the pay-off
            table.

        Note:
            The nadir point if an estimate, and therefore might be a bad
            representation of the real nadir point.
        """
        pay_off_table = np.zeros(
            (self.__problem.n_of_objectives, self.problem.n_of_objectives)
        )
        func = self._evaluator
        bounds = self.problem.get_variable_bounds()

        for ind in range(self.problem.n_of_objectives):
            args = (ind,)
            x = self.method.run(func, bounds, args)

            pay_off_table[ind], _ = self.problem.evaluate(x)

        # The ideal point can be found on the diagonal of the PO-table,
        # and an estimate of the nadir by taking the maximun value of each
        # column.
        return np.array(
            [pay_off_table.diagonal(), np.amax(pay_off_table, axis=0)]
        )



"""Various interactive multiobjective optimization methods belonging to them
NIMBUS-family are defined here

"""

import logging
import logging.config
from copy import deepcopy
from os import path
from typing import List, Optional, Tuple, Union

import numpy as np

from desdeov2.methods.InteractiveMethod import (
    InteractiveMethodBase,
    InteractiveMethodError,
)
from desdeov2.problem.Constraint import ScalarConstraint
from desdeov2.problem.Problem import (
    ProblemBase,
    ScalarDataProblem,
    ScalarMOProblem,
)
from desdeov2.solver.ASF import (
    AugmentedGuessASF,
    MaxOfTwoASF,
    PointMethodASF,
    StomASF,
)
from desdeov2.solver.NumericalMethods import DiscreteMinimizer
from desdeov2.solver.ScalarSolver import ASFScalarSolver
from desdeov2.utils.frozen import frozen

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


@frozen(logger)
class SNimbus(InteractiveMethodBase):
    """Implements the synchronous NIMBUS variant as defined in
       `Miettinen 2016`_

    Attributes:
        pareto_front (np.ndarray): The representation of the pareto front of
        the problem to be solved.
        objective_vectors (np.ndarray): The objective vector values
        corresponding to the representation of the pareto front.
        classifications (List[Tuple[str, Optional[float]]]): Tuples
        representing the classifications of each objective funtion. The second
        element of the tuple is an extra information needed by some
        classifications.
        nadir (np.ndarray): The nadir point.
        ideal (np.ndarray): The ideal point.
        cind (int): current index of the objective solution (used to map it
        back to the decision variables)
        current_point (np.ndarray): The currently selected objective vector.
        archive (List[np.ndarray]): Archived solutions.
        n_points (int): Number of points to be generated each iteration.
        first_iteration (bool): Indicated whether the method has not been
        iterated yet.
        generate_intermediate (bool): Should intermediate points be generated
        in the next iteration?
        search_between_points (Tuple[np.ndarray]): The two points between the
        intermediate points should be generated.
        n_intermediate_solutions (int): Number of intermediate solutions to be
        generated between two objective vectors.

    .. _Miettinen 2006:
        Mietttinen, K. & Mäkelä, Marko M.
        Synchronous approach in interactive multiobjective optimization
        European Journal of Operational Research, 2006, 170, 909-922

    """

    def __init__(self, problem: ProblemBase):
        super().__init__(problem)

        if isinstance(problem, ScalarMOProblem):
            msg = (
                "Currently E-NAUTILUS works only with the "
                "ScalarDataProblem problem class or by sypplying "
                "the pareto set and objectives manually."
            )
            logger.error(msg)
            raise NotImplementedError(msg)

        # The full pareto front
        self.__pareto_front: np.ndarray = None
        # The objective vector values correspond to the front
        self.__objective_vectors: np.ndarray = None
        # classifications of the objective functions
        self.__classifications: List[Tuple[str, Optional[float]]] = []
        self.__available_classifications: List[str] = [
            "<",
            "<=",
            "=",
            ">=",
            "0",
        ]
        # index sets to keep track how each objective should change
        self.__ind_set_lt: List[int] = []
        self.__ind_set_lte: List[int] = []
        self.__ind_set_eq: List[int] = []
        self.__ind_set_gte: List[int] = []
        self.__ind_set_free: List[int] = []
        # bounds
        self.__aspiration_levels: np.ndarray = None
        self.__upper_bounds: np.ndarray = None
        # nadir
        self.__nadir: np.ndarray = None
        # ideal
        self.__ideal: np.ndarray = None
        # current index of the solution used
        self.__cind: np.int = 0
        # currently selected objective vector
        self.__current_point: np.ndarray = None
        # solution archive
        self.__archive: List[Tuple[np.ndarray, np.ndarray]] = []
        # number of point to be generated
        self.__n_points: int = 0
        # flag to represent if first iteration
        self.__first_iteration: bool = True
        # flag to generate intermediate points
        self.__generate_intermediate: bool = False
        # points between intermediate solutions are explored
        self.__search_between_points: Tuple[np.ndarray, np.ndarray] = (
            None,
            None,
        )  # noqa
        # number of intermediate points to be generated
        self.__n_intermediate_solutions: int = 0
        # subproblems
        self.__sprob_1: Optional[ScalarDataProblem] = None
        self.__sprob_2: Optional[ScalarDataProblem] = None
        self.__sprob_3: Optional[ScalarDataProblem] = None
        self.__sprob_4: Optional[ScalarDataProblem] = None

        # solvers for each subproblem
        self.__solver_1: Optional[ASFScalarSolver] = None
        self.__solver_2: Optional[ASFScalarSolver] = None
        self.__solver_3: Optional[ASFScalarSolver] = None
        self.__solver_4: Optional[ASFScalarSolver] = None

    @property
    def pareto_front(self) -> np.ndarray:
        return self.__pareto_front

    @pareto_front.setter
    def pareto_front(self, val: np.ndarray):
        self.__pareto_front = val

    @property
    def objective_vectors(self) -> np.ndarray:
        return self.__objective_vectors

    @objective_vectors.setter
    def objective_vectors(self, val: np.ndarray):
        self.__objective_vectors = val

    @property
    def classifications(self) -> List[Tuple[str, Optional[float]]]:
        return self.__classifications

    @classifications.setter
    def classifications(self, val: List[Tuple[str, Optional[float]]]):
        """Parses classifications and checks if they are sensical. See
        `Miettinen 2016`_

        Args:
            val (List[Tuple, Optional[float]]): The classificaitons. The first
            element is the class and the second element is auxilliary
            information needed by some classifications.

        Raises:
            InteractiveMethodError: The classifications given are ill-formed.

        """
        if len(val) != self.objective_vectors.shape[1]:
            msg = (
                "Each of the objective functions must be classified. Check "
                "that '{}' has a correct amount (in this case {}) of "
                "elements in it."
            ).format(val, self.objective_vectors.shape[1])
            logger.error(msg)
            raise InteractiveMethodError(msg)

        if not all(
            [cls[0] in self.__available_classifications for cls in val]
        ):
            msg = (
                "Check the given classifications '{}'. The first element of "
                "each tuple should be found in '{}'"
            ).format(val, self.__available_classifications)
            logger.error(msg)
            raise InteractiveMethodError(msg)

        clss = [cls[0] for cls in val]
        if not (
            ("<" in clss or "<=" in clss) and (">=" in clss or "0" in clss)
        ):
            msg = (
                "Check the calssifications '{}'. At least one of the "
                "objectives should able to be improved and one of the "
                "objectives should be able to deteriorate."
            ).format(val)
            logger.error(msg)
            raise InteractiveMethodError(msg)

        for (ind, cls) in enumerate(val):
            if cls[0] == "<=":
                if not cls[1] < self.current_point[ind]:
                    msg = (
                        "For the '{}th' objective, the aspiration level '{}' "
                        "must be smaller than the current value of the "
                        "objective '{}'."
                    ).format(ind, cls[1], self.current_point[ind])
                    logger.error(msg)
                    raise InteractiveMethodError(msg)
            elif cls[0] == ">=":
                if not cls[1] > self.current_point[ind]:
                    msg = (
                        "For the '{}th' objective, the upper bound '{}' "
                        "must be greater than the current value of the "
                        "objective '{}'."
                    ).format(ind, cls[1], self.current_point[ind])
                    logger.error(msg)
                    raise InteractiveMethodError(msg)

        self.__classifications = val

    @property
    def nadir(self) -> np.ndarray:
        return self.__nadir

    @nadir.setter
    def nadir(self, val: np.ndarray):
        self.__nadir = val

    @property
    def ideal(self) -> np.ndarray:
        return self.__ideal

    @ideal.setter
    def ideal(self, val: np.ndarray):
        self.__ideal = val

    @property
    def cind(self) -> int:
        return self.__cind

    @cind.setter
    def cind(self, val: int):
        self.__cind = val

    @property
    def archive(self) -> List[Tuple[np.ndarray, np.ndarray]]:
        return self.__archive

    @archive.setter
    def archive(self, val: List[Tuple[np.ndarray, np.ndarray]]):
        self.__archive = val

    @property
    def current_point(self) -> np.ndarray:
        return self.__current_point

    @current_point.setter
    def current_point(self, val: np.ndarray):
        """Set the current point for the algorithm. The dimensions of the
        current point must match the dimensions of the row vectors in
        objective_vectors.

        Args:
            val(np.ndarray): The current point.

        Raises:
            InteractiveMethodError: The current point's dimension does not
            match that of the given pareto optimal objective vectors.

        """
        if len(val) != self.objective_vectors.shape[1]:
            msg = (
                "Current point dimensions '{}' don't match the objective"
                " vector dimensions '{}'"
            ).format(len(val), self.objective_vectors.shape[1])
            logger.error(msg)
            raise InteractiveMethodError(msg)

        self.__current_point = val

    @property
    def n_points(self) -> int:
        return self.__n_points

    @n_points.setter
    def n_points(self, val: int):
        """Set the desired number of solutions to be generated each
        iteration. Must be between 1 and 4 (inclusive)

        Args:
            val(int): The number of points to be generated.

        Raises:
            InteractiveMethodError: The number of points to be generated is not
            between 1 and 4.

        """
        if val < 1 or val > 4:
            msg = (
                "The given number '{}' of solutions to be generated is not"
                " between 1 and 4 (inclusive)"
            ).format(val)
            logger.error(msg)
            raise InteractiveMethodError(msg)
        self.__n_points = val

    @property
    def first_iteration(self) -> bool:
        return self.__first_iteration

    @first_iteration.setter
    def first_iteration(self, val: bool):
        self.__first_iteration = val

    @property
    def generate_intermediate(self) -> bool:
        return self.__generate_intermediate

    @generate_intermediate.setter
    def generate_intermediate(self, val: bool):
        self.__generate_intermediate = val

    @property
    def search_between_points(self) -> Tuple[np.ndarray, np.ndarray]:
        return self.__search_between_points

    @search_between_points.setter
    def search_between_points(self, val: Tuple[np.ndarray, np.ndarray]):
        if len(val) != 2:
            msg = (
                "To generate intermediate points, two points must be "
                "specified. Number of given points were {}"
            ).format(len(val))
            logger.error(msg)
            raise InteractiveMethodError(msg)
        self.__search_between_points = val

    @property
    def n_intermediate_solutions(self) -> int:
        return self.__n_intermediate_solutions

    @n_intermediate_solutions.setter
    def n_intermediate_solutions(self, val: int):
        if val < 1:
            msg = (
                "Number of intermediate points to be generated must be "
                "positive definitive. Given number of points {}."
            ).format(val)
            logger.error(msg)
            raise InteractiveMethodError(msg)

        self.__n_intermediate_solutions = val

    def _sort_classsifications(self):
        """Sort the objective indices into their corresponding sets and save
        the aspiration and upper bounds set in the classifications.

        Raises:
            InteractiveMethodError: A classification is found to be ill-formed.

        """
        # empty the sets and bounds
        self.__ind_set_lt = []
        self.__ind_set_lte = []
        self.__ind_set_eq = []
        self.__ind_set_gte = []
        self.__ind_set_free = []
        aspiration_levels = []
        upper_bounds = []
        for (ind, cls) in enumerate(self.classifications):
            if cls[0] == "<":
                self.__ind_set_lt.append(ind)
            elif cls[0] == "<=":
                self.__ind_set_lte.append(ind)
                aspiration_levels.append(cls[1])
            elif cls[0] == "=":
                self.__ind_set_eq.append(ind)
            elif cls[0] == ">=":
                self.__ind_set_gte.append(ind)
                upper_bounds.append(cls[1])
            elif cls[0] == "0":
                self.__ind_set_free.append(ind)
            else:
                msg = (
                    "Check that the classification '{}' is correct."
                ).format(cls)
                logger.error(msg)
                raise InteractiveMethodError(msg)
        self.__aspiration_levels = np.array(aspiration_levels)
        self.__upper_bounds = np.array(upper_bounds)

    def _create_reference_point(self) -> np.ndarray:
        """Create a reference point indicating the DM's preferences using the
        classifications given by the DM.

        Returns:
            np.ndarray: The reference point.

        """
        ref_point = np.zeros(self.objective_vectors.shape[1])
        ref_point[self.__ind_set_lt] = self.ideal[self.__ind_set_lt]
        ref_point[self.__ind_set_lte] = self.__aspiration_levels
        ref_point[self.__ind_set_eq] = self.current_point[self.__ind_set_eq]
        ref_point[self.__ind_set_gte] = self.__upper_bounds
        ref_point[self.__ind_set_free] = self.nadir[self.__ind_set_free]

        return ref_point

    def _create_intermediate_reference_points(self) -> np.ndarray:
        f1 = self.search_between_points[0]
        f2 = self.search_between_points[1]
        points = []

        # The unnormalized vector pointing from f1 to f2
        normal = f2 - f1

        # fraction of the normal
        step_vec = normal / (self.n_intermediate_solutions + 1)
        points.append(f1 + step_vec)

        for i in range(1, self.n_intermediate_solutions):
            points.append(points[i - 1] + step_vec)

        return np.array(points)

    def initialize(  # type: ignore
        self, n_solutions: int, starting_point: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Initialize the method and return the starting objective vector.

        Args:
            n_solutions(int): The number of solutions to be generated each
            iteration.
            starting_point(Optional[np.ndarray]): An objective vector to
            function as the starting point for synchronous NIMBUS.
            pareto_front(Optional[np.ndarray]): A 2D array representing the
            pareto optimal solutions of the problem to be solved.
            objective_vectors(Optional[np.ndarray]): The objective vectors
            corresponding to each pareto optimal solution.

        Returns:
            np.ndarray: The starting point of the algorithm.

        Raises:
            NotImplementedError: The problem is something else than
            ScalarDataProblem

        Note:
            The data to be used must be available in the underlying problem OR
            given explicitly.

        """
        self.n_points = n_solutions

        if isinstance(self.problem, ScalarDataProblem):
            self.pareto_front = self.problem.decision_vectors
            self.objective_vectors = self.problem.objective_vectors

        else:
            msg = "Only supoorts solving for SacalarDataProblem at the moment."
            logger.error(msg)
            raise NotImplementedError(msg)

        self.nadir = np.max(self.objective_vectors, axis=0)
        self.ideal = np.min(self.objective_vectors, axis=0)

        self.__classifications = [("", None)] * self.objective_vectors.shape[1]

        # if the starting point has been specified, use that. Otherwise, use
        # random one.
        if starting_point is not None:
            self.current_point = starting_point

        else:
            rind = np.random.randint(0, len(self.objective_vectors))
            self.current_point = self.objective_vectors[rind]

        return self.current_point

    def iterate(
        self
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray]]:
        """Iterate according to the preferences given by the DM in the
        interaction phase.

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray]]:
            Returns the current point for the first iteration. For the
            following iterations, returns the decision vectors, the objective
            vectors values and the current archive of saved points.

        """
        # if first iteration, just return the starting point
        if self.first_iteration:
            self.first_iteration = False
            return self.current_point

        res_all_xs: List[np.ndarray] = []
        res_all_fs: List[np.ndarray] = []

        if self.generate_intermediate:
            # generate n points between two previous points
            # can reuse the solver for subrpoblem 3 here
            if self.__sprob_3 is None:
                # create the subproblem and solver
                self.__sprob_3 = ScalarDataProblem(
                    self.pareto_front, self.objective_vectors
                )
                self.__sprob_3.nadir = self.nadir
                self.__sprob_3.ideal = self.ideal

                self.__solver_3 = ASFScalarSolver(
                    self.__sprob_3, DiscreteMinimizer()
                )
                self.__solver_3.asf = PointMethodASF(self.nadir, self.ideal)

            z_bars = self._create_intermediate_reference_points()
            for z in z_bars:
                res = self.__solver_3.solve(z)  # type: ignore
                res_all_xs.append(res[0])
                res_all_fs.append(res[1][0])

            # always require and explicit request from the DM to generate
            # intermediate points
            self.generate_intermediate = False

        else:
            # solve the subproblems normally
            if self.n_points >= 1:
                # solve the desired number of ASFs
                self._sort_classsifications()
                z_bar = self._create_reference_point()

                # subproblem 1
                if self.__sprob_1 is None:
                    # create the subproblem and solver
                    self.__sprob_1 = ScalarDataProblem(
                        self.pareto_front, self.objective_vectors
                    )
                    self.__sprob_1.nadir = self.nadir
                    self.__sprob_1.ideal = self.ideal

                    self.__solver_1 = ASFScalarSolver(
                        self.__sprob_1, DiscreteMinimizer()
                    )
                    self.__solver_1.asf = MaxOfTwoASF(
                        self.nadir, self.ideal, [], []
                    )

                # set the constraints for the 1st subproblem
                sp1_all_cons = []
                sp1_cons1_idx = np.sort(
                    (
                        self.__ind_set_lt
                        + self.__ind_set_lte
                        + self.__ind_set_eq
                    )
                )

                if len(sp1_cons1_idx) > 0:
                    sp1_cons1_f = lambda _, fs: np.where(  # noqa
                        np.all(
                            fs[:, sp1_cons1_idx]
                            <= self.current_point[sp1_cons1_idx],
                            axis=1,
                        ),
                        np.ones(len(fs)),
                        -np.ones(len(fs)),
                    )
                    sp1_cons1 = ScalarConstraint(
                        "sp1_cons1",
                        self.pareto_front.shape[1],
                        self.objective_vectors.shape[1],
                        sp1_cons1_f,
                    )
                    sp1_all_cons.append(sp1_cons1)

                sp1_cons2_idx = self.__ind_set_gte
                if len(sp1_cons2_idx) > 0:
                    sp1_cons2_f = lambda _, fs: np.where(  # noqa
                        np.all(
                            fs[:, sp1_cons2_idx] <= self.__upper_bounds, axis=1
                        ),
                        np.ones(len(fs)),
                        -np.ones(len(fs)),
                    )
                    sp1_cons2 = ScalarConstraint(
                        "sp1_cons2",
                        self.pareto_front.shape[1],
                        self.objective_vectors.shape[1],
                        sp1_cons2_f,
                    )
                    sp1_all_cons.append(sp1_cons2)

                self.__sprob_1.constraints = sp1_all_cons

                # solve the subproblem
                self.__solver_1.asf.lt_inds = self.__ind_set_lt  # type: ignore
                self.__solver_1.asf.lte_inds = (  # type: ignore
                    self.__ind_set_lte
                )  # type: ignore, # noqa

                sp1_reference = np.zeros(self.objective_vectors.shape[1])
                sp1_reference[self.__ind_set_lte] = self.__aspiration_levels

                res1 = self.__solver_1.solve(sp1_reference)  # type: ignore
                res_all_xs.append(res1[0])
                res_all_fs.append(res1[1][0])

            if self.n_points >= 2:
                # subproblem 2
                if self.__sprob_2 is None:
                    # create the subproblem and solver
                    self.__sprob_2 = ScalarDataProblem(
                        self.pareto_front, self.objective_vectors
                    )
                    self.__sprob_2.nadir = self.nadir
                    self.__sprob_2.ideal = self.ideal

                    self.__solver_2 = ASFScalarSolver(
                        self.__sprob_2, DiscreteMinimizer()
                    )
                    self.__solver_2.asf = StomASF(self.ideal)

                res2 = self.__solver_2.solve(z_bar)  # type: ignore
                res_all_xs.append(res2[0])
                res_all_fs.append(res2[1][0])

            if self.n_points >= 3:
                # subproblem 3
                if self.__sprob_3 is None:
                    # create the subproblem and solver
                    self.__sprob_3 = ScalarDataProblem(
                        self.pareto_front, self.objective_vectors
                    )
                    self.__sprob_3.nadir = self.nadir
                    self.__sprob_3.ideal = self.ideal

                    self.__solver_3 = ASFScalarSolver(
                        self.__sprob_3, DiscreteMinimizer()
                    )
                    self.__solver_3.asf = PointMethodASF(
                        self.nadir, self.ideal
                    )

                res3 = self.__solver_3.solve(z_bar)  # type: ignore
                res_all_xs.append(res3[0])
                res_all_fs.append(res3[1][0])

            if self.n_points >= 4:
                # subproblem 4
                if self.__sprob_4 is None:
                    # create the subproblem and solver
                    self.__sprob_4 = ScalarDataProblem(
                        self.pareto_front, self.objective_vectors
                    )
                    self.__sprob_4.nadir = self.nadir
                    self.__sprob_4.ideal = self.ideal

                    self.__solver_4 = ASFScalarSolver(
                        self.__sprob_4, DiscreteMinimizer()
                    )
                    self.__solver_4.asf = AugmentedGuessASF(
                        self.nadir, self.ideal, self.__ind_set_free
                    )
                else:
                    # update the indices to be excluded in the existing
                    # solver's asf
                    self.__solver_4.asf.indx_to_exclude = (  # type: ignore
                        self.__ind_set_free
                    )

                res4 = self.__solver_4.solve(z_bar)  # type: ignore
                res_all_xs.append(res4[0])
                res_all_fs.append(res4[1][0])

        # return the obtained solutions and the archiveof existing solutions
        # deepcopy, beacuse we dont want to return a reference to the archive
        return (
            np.array(res_all_xs),
            np.array(res_all_fs),
            deepcopy(self.archive),
        )

    def interact(
        self,
        most_preferred_point: Optional[np.ndarray] = None,
        classifications: Optional[List[Tuple[str, Optional[float]]]] = None,
        n_generated_solutions: int = -1,
        save_points: Optional[List[Tuple[np.ndarray, np.ndarray]]] = None,
        search_between_points: Optional[Tuple[np.ndarray]] = None,
        n_intermediate_solutions: int = 1,
    ):
        """Handle the the preferneces given by the DM and set up variables for
        the next iterations

        Args:
            most_preferred_point (Optional[np.ndarray]): The most preferred
            point.
            classificaitons (Optional[List[Tuple[str, Optional[float]]]]): The
            classifications of the objective functions.
            n_generated_solutions (Optional[int]): The number of solutions to
            be generated in the next iteration.
            save_points (Optional[List[Tuple[np.ndarray, np.ndarray]]]): A list
            of points to be saved in the archive.
            search_between_points (Optional[Tuple[np.ndarray]]): The two points
            between intermediate solutions are searched for.
            n_intermediate_solutions (int): The number of intermediate
            solutions generated between the two given points.

        """
        if most_preferred_point is not None:
            self.current_point = most_preferred_point

        if classifications is not None:
            # parse new classificaitons and sort the indices and given
            # prefernece in an usable form
            self.classifications = classifications

        if n_generated_solutions > 0:
            # change the number of subrpoblems to be solved in the next
            # iteration
            self.n_points = n_generated_solutions

        if save_points is not None:
            # save the given points in the archive
            self.archive.extend(save_points)

        if search_between_points is not None:
            # set the flag indicating a preference to generate intermediate
            # points in the next iteration
            self.generate_intermediate = True
            self.search_between_points = search_between_points  # type: ignore
            self.n_intermediate_solutions = n_intermediate_solutions



"""Base classes for interactive methods.

"""

import abc
from abc import abstractmethod

from desdeov2.problem.Problem import ProblemBase


class InteractiveMethodError(Exception):
    """Raised when an error is encountered in the InteractiveMethod classes.

    """

    pass


class InteractiveMethodBase(abc.ABC):
    """Defines a common interface to be used by all the interactive MOO
    methods.

    Arguments:
        problem(ProblemBase): The given problem to be solved using an
        interactive method.

    Attributes:
        problem(ProblemBase): The given problem to be solved using an
        interactive method.

    """

    def __init__(self, problem: ProblemBase):
        self.__problem: ProblemBase = problem

    @property
    def problem(self) -> ProblemBase:
        return self.__problem

    @problem.setter
    def problem(self, val: ProblemBase):
        self.__problem = val

    @abstractmethod
    def initialize(self):
        """Initializes the method.

        """
        pass

    @abstractmethod
    def iterate(self):
        """Iterate the method once.

        """
        pass

    @abstractmethod
    def interact(self):
        """Change the state of the method given a decision maker's
        preference.

        """
        pass



"""Here various classes are defined that represent multiobjective optimization
problems.

"""

import logging
import logging.config
from abc import ABC, abstractmethod
from os import path
from typing import List, Optional, Tuple, Union

import numpy as np

from desdeov2.problem.Constraint import ScalarConstraint
from desdeov2.problem.Objective import ScalarObjective
from desdeov2.problem.Variable import Variable

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


class ProblemError(Exception):
    """Raised when an error related to the Problem class is encountered.

    """


class ProblemBase(ABC):
    """The base class from which every other class representing a problem should
    derive.

    """

    def __init__(self):
        self.__nadir: np.ndarray = None
        self.__ideal: np.ndarray = None
        self.__n_of_objectives: int = 0
        self.__n_of_variables: int = 0
        self.__decision_vectors: np.ndarray = None
        self.__objective_vectors: np.ndarray = None

    @property
    def nadir(self) -> np.ndarray:
        return self.__nadir

    @nadir.setter
    def nadir(self, val: np.ndarray):
        self.__nadir = val

    @property
    def ideal(self) -> np.ndarray:
        return self.__ideal

    @ideal.setter
    def ideal(self, val: np.ndarray):
        self.__ideal = val

    @property
    def n_of_objectives(self) -> int:
        return self.__n_of_objectives

    @n_of_objectives.setter
    def n_of_objectives(self, val: int):
        self.__n_of_objectives = val

    @property
    def n_of_variables(self) -> int:
        return self.__n_of_variables

    @n_of_variables.setter
    def n_of_variables(self, val: int):
        self.__n_of_variables = val

    @property
    def decision_vectors(self) -> np.ndarray:
        return self.__decision_vectors

    @decision_vectors.setter
    def decision_vectors(self, val: np.ndarray):
        self.__decision_vectors = val

    @property
    def objective_vectors(self) -> np.ndarray:
        return self.__objective_vectors

    @objective_vectors.setter
    def objective_vectors(self, val: np.ndarray):
        self.__objective_vectors = val

    @abstractmethod
    def get_variable_bounds(self) -> Union[None, np.ndarray]:
        pass

    @abstractmethod
    def evaluate(
        self, decision_vectors: np.ndarray
    ) -> Union[np.ndarray, Tuple[np.ndarray, Union[np.ndarray, None]]]:
        """Evaluates the problem using an ensemble of input vectors.

        Args:
            decision_vectors (np.ndarray): An array of decision variable
            input vectors.

        Returns:
            (tuple): tuple containing:
                solutions (np.ndarray): The corresponding objective function
                values for each input vector.
                constraints (Union[np.ndarray, None]): The constraint values
                of the problem corresponding each input vector.

        """
        pass

    @abstractmethod
    def evaluate_constraint_values(self) -> Optional[np.ndarray]:
        """Evaluate just the constraint function values using the attributes
        decision_vectors and objective_vectors

        Note:
            Currently not supported by ScalarMOProblem

        """
        pass


class ScalarMOProblem(ProblemBase):
    """A multiobjective optimization problem with user defined objective funcitons,
    constraints and variables. The objectives each return a single scalar.

    Args:
        objectives (List[ScalarObjective]): A list containing the objectives of
        the problem.
        variables (List[Variable]): A list containing the variables of the
        problem.
        constraints (List[ScalarConstraint]): A list containing the
        constraints of the problem. If no constraints exist, None may
        be supllied as the value.
        nadir (Optional[np.ndarray]): The nadir point of the problem.
        ideal (Optional[np.ndarray]): The ideal point of the problem.

    Attributes:
        n_of_objectives (int): The number of objectives in the problem.
        n_of_variables (int): The number of variables in the problem.
        n_of_constraints (int): The number of constraints in the problem.
        nadir (np.ndarray): The nadir point of the problem.
        ideal (np.ndarray): The ideal point of the problem.
        objectives (List[ScalarObjective]): A list containing the objectives of
        the problem.
        constraints (List[ScalarConstraint]): A list conatining the constraints
        of the problem.

    Raises:
        ProblemError: Ill formed nadir and/or ideal vectors are supplied.

    """

    def __init__(
        self,
        objectives: List[ScalarObjective],
        variables: List[Variable],
        constraints: List[ScalarConstraint],
        nadir: Optional[np.ndarray] = None,
        ideal: Optional[np.ndarray] = None,
    ) -> None:
        super().__init__()
        self.__objectives: List[ScalarObjective] = objectives
        self.__variables: List[Variable] = variables
        self.__constraints: List[ScalarConstraint] = constraints

        self.__n_of_objectives: int = len(self.objectives)
        self.__n_of_variables: int = len(self.variables)

        if self.constraints is not None:
            self.__n_of_constraints: int = len(self.constraints)
        else:
            self.__n_of_constraints = 0

        # Nadir vector must be the same size as the number of objectives
        if nadir is not None:
            if len(nadir) != self.n_of_objectives:
                msg = (
                    "The length of the nadir vector does not match the"
                    "number of objectives: Length nadir {}, number of "
                    "objectives {}."
                ).format(len(nadir), self.n_of_objectives)
                logger.error(msg)
                raise ProblemError(msg)

        # Ideal vector must be the same size as the number of objectives
        if ideal is not None:
            if len(ideal) != self.n_of_objectives:
                msg = (
                    "The length of the ideal vector does not match the"
                    "number of objectives: Length ideal {}, number of "
                    "objectives {}."
                ).format(len(ideal), self.n_of_objectives)
                logger.error(msg)
                raise ProblemError(msg)

        # Nadir and ideal vectors must match in size
        if nadir is not None and ideal is not None:
            if len(nadir) != len(ideal):
                msg = (
                    "The length of the nadir and ideal point don't match:"
                    " length of nadir {}, length of ideal {}."
                ).format(len(nadir), len(ideal))
                logger.error(msg)
                raise ProblemError(msg)

        self.__nadir = nadir
        self.__ideal = ideal

    @property
    def n_of_constraints(self) -> int:
        return self.__n_of_constraints

    @n_of_constraints.setter
    def n_of_constraints(self, val: int):
        self.__n_of_constraints = val

    @property
    def objectives(self) -> List[ScalarObjective]:
        return self.__objectives

    @objectives.setter
    def objectives(self, val: List[ScalarObjective]):
        self.__objectives = val

    @property
    def variables(self) -> List[Variable]:
        return self.__variables

    @variables.setter
    def variables(self, val: List[Variable]):
        self.__variables = val

    @property
    def constraints(self) -> List[ScalarConstraint]:
        return self.__constraints

    @constraints.setter
    def constraints(self, val: List[ScalarConstraint]):
        self.__constraints = val

    @property
    def n_of_constraints(self) -> int:
        return self.__n_of_constraints

    @n_of_constraints.setter
    def n_of_constraints(self, val: int):
        self.__n_of_constraints = val

    @property
    def n_of_objectives(self) -> int:
        return self.__n_of_objectives

    @n_of_objectives.setter
    def n_of_objectives(self, val: int):
        self.__n_of_objectives = val

    @property
    def n_of_variables(self) -> int:
        return self.__n_of_variables

    @n_of_variables.setter
    def n_of_variables(self, val: int):
        self.__n_of_variables = val

    @property
    def nadir(self) -> np.ndarray:
        return self.__nadir

    @nadir.setter
    def nadir(self, val: np.ndarray):
        self.__nadir = val

    @property
    def ideal(self) -> np.ndarray:
        return self.__ideal

    @ideal.setter
    def ideal(self, val: np.ndarray):
        self.__ideal = val

    def get_variable_bounds(self) -> Union[np.ndarray, None]:
        """Return the upper and lower bounds of each decision variable present
        in the problem as a 2D numpy array. The first column corresponds to the
        lower bounds of each variable, and the second column to the upper
        bound.

        Returns:
           np.ndarray: Lower and upper bounds of each variable
           as a 2D numpy array. If undefined variables, return None instead.

        """
        if self.variables is not None:
            bounds = np.ndarray((self.n_of_variables, 2))
            for ind, var in enumerate(self.variables):
                bounds[ind] = np.array(var.get_bounds())
            return bounds
        else:
            logger.info(
                "Attempted to get variable bounds for a "
                "ScalarMOProblem with no defined variables."
            )
            return None

    def get_variable_names(self) -> List[str]:
        """Return the variable names of the variables present in the problem in
        the order they were added.

        Returns:
            List[str]: Names of the variables in the order they were added.

        """
        return [var.name for var in self.variables]

    def get_objective_names(self) -> List[str]:
        """Return the names of the objectives present in the problem in the
        order they were added.

        Returns:
            List[str]: Names of the objectives in the order they were added.

        """
        return [obj.name for obj in self.objectives]

    def get_variable_lower_bounds(self) -> np.ndarray:
        """Return the lower bounds of each variable as a list. The order of the bounds
        follows the order the variables were added to the problem.

        Returns:
            np.ndarray: An array with the lower bounds of the variables.
        """
        return np.array([var.get_bounds()[0] for var in self.variables])

    def get_variable_upper_bounds(self) -> np.ndarray:
        """Return the upper bounds of each variable as a list. The order of the bounds
        follows the order the variables were added to the problem.

        Returns:
            np.ndarray: An array with the upper bounds of the variables.
        """
        return np.array([var.get_bounds()[1] for var in self.variables])

    def evaluate(
        self, decision_vectors: np.ndarray
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """Evaluates the problem using an ensemble of input vectors.

        Args:
            decision_vectors (np.ndarray): An 2D array of decision variable
            input vectors. Each column represent the values of each decision
            variable.

        Returns:
            Tuple[np.ndarray, Union[None, np.ndarray]]: If constraint are
            defined, returns the objective vector values and corresponding
            constraint values. Or, if no constraints are defined, returns just
            the objective vector values with None as the constraint values.

        Raises:
            ProblemError: The decision_vectors have wrong dimensions.

        """
        # Reshape decision_vectors with single row to work with the code
        shape = np.shape(decision_vectors)
        if len(shape) == 1:
            decision_vectors = np.reshape(decision_vectors, (1, shape[0]))

        (n_rows, n_cols) = np.shape(decision_vectors)

        if n_cols != self.n_of_variables:
            msg = (
                "The length of the input vectors does not match the number "
                "of variables in the problem: Input vector length {}, "
                "number of variables {}."
            ).format(n_cols, self.n_of_variables)
            logger.error(msg)
            raise ProblemError(msg)

        objective_vectors: np.ndarray = np.ndarray(
            (n_rows, self.n_of_objectives), dtype=float
        )
        if self.n_of_constraints > 0:
            constraint_values: np.ndarray = np.ndarray(
                (n_rows, self.n_of_constraints), dtype=float
            )
        else:
            constraint_values = None

        # Calculate the objective values
        for (col_i, objective) in enumerate(self.objectives):
            objective_vectors[:, col_i] = np.array(
                list(map(objective.evaluate, decision_vectors))
            )

        # Calculate the constraint values
        if constraint_values is not None:
            for (col_i, constraint) in enumerate(self.constraints):
                constraint_values[:, col_i] = np.array(
                    list(
                        map(
                            constraint.evaluate,
                            decision_vectors,
                            objective_vectors,
                        )
                    )
                )

        return (objective_vectors, constraint_values)

    def evaluate_constraint_values(self) -> Optional[np.ndarray]:
        """Evaluate just the constraint function values using the attributes
        decision_vectors and objective_vectors

        Raises:
            NotImplementedError

        Note:
            Currently not supported by ScalarMOProblem

        """
        raise NotImplementedError("Not implemented for ScalarMOProblem")


class ScalarDataProblem(ProblemBase):
    """Defines a problem with pre-computed data representing a multiobjective
    optimization problem with scalar valued objective functions.

    Args:
        decision_vectors (np.ndarray): A 2D vector of decision_vectors. Each
        row represents a solution with the value for each decision_vectors
        defined on the columns.
        objective_vectors (np.ndarray): A 2D vector of
        objective function values. Each row represents one objective vector
        with the values for the invidual objective functions defined on the
        columns.

    Attributes:
        decision_vectors (np.ndarray): See args
        objective_vectors (np.ndarray): See args
        epsilon (float): A small floating point number to shift the bounds of
        the variables. See, get_variable_bounds
        constraints (List[ScalarConstraint]): A list of defined constraints.
        nadir (np.ndarray): The nadir point of the problem.
        ideal (np.ndarray): The ideal point of the problem.

    Note:
        It is assumed that the decision_vectors and objectives follow a direct
        one-to-one mapping, i.e., the objective values on the ith row in
        'objectives' should represent the solution of the multiobjective
        problem when evaluated with the decision_vectors on the ith row in
        'decision_vectors'.

    """

    def __init__(
        self, decision_vectors: np.ndarray, objective_vectors: np.ndarray
    ):
        super().__init__()
        self.decision_vectors: np.ndarray = decision_vectors
        self.objective_vectors: np.ndarray = objective_vectors
        # epsilon is used when computing the bounds. We don't want to exclude
        # any of the solutions that contain border values.
        # See get_variable_bounds
        self.__epsilon: float = 1e-6
        # Used to indicate if a model has been built to represent the model.
        # Used in the evaluation.
        self.__model_exists: bool = False
        self.__constraints: List[ScalarConstraint] = []

        try:
            self.n_of_variables = self.decision_vectors.shape[1]
        except IndexError as e:
            msg = (
                "Check the variable dimensions. Is it a 2D array? "
                "Encountered '{}'".format(str(e))
            )
            logger.error(msg)
            raise ProblemError(msg)

        try:
            self.n_of_objectives = self.objective_vectors.shape[1]
        except IndexError as e:
            msg = (
                "Check the objective dimensions. Is it a 2D array? "
                "Encountered '{}'".format(str(e))
            )
            logger.error(msg)
            raise ProblemError(msg)

        self.nadir = np.max(self.objective_vectors, axis=0)
        self.ideal = np.min(self.objective_vectors, axis=0)

    @property
    def epsilon(self) -> float:
        return self.__epsilon

    @epsilon.setter
    def epsilon(self, val: float):
        self.__epsilon = val

    @property
    def constraints(self) -> List[ScalarConstraint]:
        return self.__constraints

    @constraints.setter
    def constraints(self, val: List[ScalarConstraint]):
        self.__constraints = val

    def get_variable_bounds(self):
        """Return the variable bounds. A small value might be added to the
        upper bounds and substracted from the lower bounds to return closed
        bounds.

        Note:
            If self.epsilon is zero, the bounds will represent an open range.

        """
        return np.stack(
            (
                np.min(self.decision_vectors, axis=0) - self.epsilon,
                np.max(self.decision_vectors, axis=0) + self.epsilon,
            ),
            axis=1,
        )

    def evaluate_constraint_values(self) -> Optional[np.ndarray]:
        """Evaluate the constraint values for each defined constraint. A
        positive value indicates that a constraint is adhered to, a negative
        value indicates a violated constraint.

        Returns:
            Optional[np.ndarray]: A 2D array with each row representing the
            constraint values for different objective vectors. One column for
            each constraint. If no constraint function are defined, returns
            None.

        """
        if len(self.constraints) == 0:
            return None

        constraint_values = np.zeros(
            (len(self.objective_vectors), len(self.constraints))
        )

        for ind, con in enumerate(self.constraints):
            constraint_values[:, ind] = con.evaluate(
                self.decision_vectors, self.objective_vectors
            )

        return constraint_values

    def evaluate(self, decision_vectors: np.ndarray) -> np.ndarray:
        """Evaluate the values of the objectives corresponding to the decision
        decision_vectors.

        Args:
            decision_vectors (np.ndarray): A 2D array with the decision
            decision_vectors to be evaluated on each row.

        Returns:
            nd.ndarray: A 2D array with the objective values corresponding to
            each decision vectors on the rows.

        Note:
            At the moment, this function just maps the given decision
            decision_vectors to the closest decision variable present (using an
            L2 distance) in the problem and returns the corresponsing objective
            vector.

            """
        if not self.__model_exists:
            logger.warning(
                "Warning: Approximating the closest known point in "
                "a data based problem. Consider building a model "
                "first (NOT IMPLEMENTED)"
            )
            idx = np.unravel_index(
                np.linalg.norm(
                    self.decision_vectors - decision_vectors, axis=1
                ).argmin(),
                self.objective_vectors.shape,
                order="F",
            )[0]

        else:
            msg = "Models not implemented yet for data based problems."
            logger.error(msg)
            raise NotImplementedError(msg)

        return (self.objective_vectors[idx],)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['desdeov2',
 'desdeov2.manager',
 'desdeov2.methods',
 'desdeov2.problem',
 'desdeov2.solver',
 'desdeov2.utils']

package_data = \
{'': ['*']}

install_requires = \
['jupyter>=1.0,<2.0',
 'numpy>=1.17,<2.0',
 'plotly>=4.1,<5.0',
 'scipy>=1.3,<2.0',
 'sklearn>=0.0.0,<0.0.1']

setup_kwargs = {
    'name': 'desdeov2',
    'version': '0.1.1',
    'description': 'Rewrite of the original DESDEO framework. Contains traditional MCDM algorithms and scalarization utilities. To be split into multiple modules.',
    'long_description': None,
    'author': 'Giovanni Misitano',
    'author_email': 'gialmisi@student.jyu.fi',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Define a general class and implementations of achievement scalarizing
functions.

"""
import abc
import logging
import logging.config
from abc import abstractmethod
from os import path
from typing import List, Union

import numpy as np

log_conf_path = path.join(
    path.dirname(path.abspath(__file__)), "../logger.cfg"
)
logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
logger = logging.getLogger(__file__)


class ASFError(Exception):
    """Raised when an error related to the ASF classes is encountered.

    """


class ASFBase(abc.ABC):
    """A base class for representing achievement scalarizing functions.
    Instances of the implementations of this class should function as
    function.

    """

    @abstractmethod
    def __call__(
        self, objective_vector: np.ndarray, reference_point: np.ndarray
    ) -> Union[float, np.ndarray]:
        """Evaluate the ASF.

        Args:
            objective_vectors (np.ndarray): The objective vectors to calulate
            the values.
            reference_point (np.ndarray): The reference point to calculate the
            values.

        Returns:
            Union[float, np.ndarray]: Either a single ASF value or a vector of
            values if objective is a 2D array.

        Note:
            The reference point may not always necessarely be feasible, but
            it's dimensions should match that of the objective vector.
        """
        pass


class SimpleASF(ASFBase):
    """Implements a simple order-representing ASF.

    Args:
        weights (np.ndarray): A weight vector that holds weights. It's
        length should match the number of objectives in the underlying
        MOO problem the achievement problem aims to solve.

    Attributes:
        weights (np.ndarray): A weight vector that holds weights. It's
        length should match the number of objectives in the underlying
        MOO problem the achievement problem aims to solve.

    """

    def __init__(self, weights: np.ndarray):
        self.__weights = weights

    @property
    def weights(self) -> np.ndarray:
        return self.__weights

    @weights.setter
    def weights(self, val: np.ndarray):
        self.__weights = val

    def __call__(
        self, objective_vector: np.ndarray, reference_point: np.ndarray
    ) -> Union[float, np.ndarray]:
        """Evaluate the simple order-representing ASF.

        Args:
            objective_vector (np.ndarray): A vector representing a solution in
            the solution space.
            reference_point (np.ndarray): A vector representing a reference
            point in the solution space.

        Raises:
            ASFError: The dimensions of the objective vector and reference
            point don't match.

        Note:
            The shaped of objective_vector and reference_point must match.

        """
        if not objective_vector.shape == reference_point.shape:
            msg = (
                "The dimensions of the objective vector {} and "
                "reference_point {} do not match."
            ).format(objective_vector, reference_point)
            logger.debug(msg)
            raise ASFError(msg)

        return np.max(
            np.where(
                np.isnan(reference_point),
                -np.inf,
                self.weights * (objective_vector - reference_point),
            )
        )


class ReferencePointASF(ASFBase):
    """Uses a reference point q and preferenial factors to scalarize a MOO problem.
    Defined in `Miettinen 2010`_ equation (2).

    Args:
        preferential_factors (np.ndarray): The preferential factors.
        nadir (np.ndarray): The nadir point of the MOO problem to be
        scalarized.
        utopian_point (np.ndarray): The utopian point of the MOO problem to be
        scalarized.
        rho (float): A small number to be used to scale the sm factor in the
        ASF. Defaults to 0.1.

    Attributes:
        preferential_factors (np.ndarray): The preferential factors.
        nadir (np.ndarray): The nadir point of the MOO problem to be
        scalarized.
        utopian_point (np.ndarray): The utopian point of the MOO problem to be
        scalarized.
        rho (float): A small number to be used to scale the sm factor in the
        ASF. Defaults to 0.1.

    .. _Miettinen 2010:
        Miettinen, K.; Eskelinen, P.; Ruiz, F. & Luque, M.
        NAUTILUS method: An interactive technique in multiobjective
        optimization based on the nadir point
        Europen Joural of Operational Research, 2010, 206, 426-434

    """

    def __init__(
        self,
        preferential_factors: np.ndarray,
        nadir: np.ndarray,
        utopian_point: np.ndarray,
        rho: float = 0.1,
    ):
        self.__preferential_factors = preferential_factors
        self.__nadir = nadir
        self.__utopian_point = utopian_point
        self.__rho = rho

    @property
    def preferential_factors(self) -> np.ndarray:
        return self.__preferential_factors

    @preferential_factors.setter
    def preferential_factors(self, val: np.ndarray):
        self.__preferential_factors = val

    @property
    def nadir(self) -> np.ndarray:
        return self.__nadir

    @nadir.setter
    def nadir(self, val: np.ndarray):
        self.__nadir = val

    @property
    def utopian_point(self) -> np.ndarray:
        return self.__utopian_point

    @utopian_point.setter
    def utopian_point(self, val: np.ndarray):
        self.__utopian_point = val

    @property
    def rho(self) -> float:
        return self.__rho

    @rho.setter
    def rho(self, val: float):
        self.__rho = val

    def __call__(
        self, objective_vector: np.ndarray, reference_point: np.ndarray
    ) -> Union[float, np.ndarray]:
        mu = self.__preferential_factors
        f = objective_vector
        q = reference_point
        rho = self.__rho
        z_nad = self.__nadir
        z_uto = self.__utopian_point

        max_term = np.max(mu * (f - q), axis=-1)
        sum_term = rho * np.sum((f - q) / (z_nad - z_uto), axis=-1)

        return max_term + sum_term


class MaxOfTwoASF(ASFBase):
    """Implements the ASF as defined in eq. 3.1 `Miettinen 2006`_

    Args:
        nadir (np.ndarray): The nadir point.
        ideal (np.ndarray): The ideal point.
        lt_inds (List[int]): Indices of the objectives categorized to be
        decreased.
        lte_inds (List[int]): Indices of the objectives categorized to be
        reduced until some value is reached.
        rho (float): A small number to form the utopian point.
        rho_sum (float): A small number to be used as a weight for the sum
        term.

    Attributes:
        nadir (np.ndarray): The nadir point.
        ideal (np.ndarray): The ideal point.
        lt_inds (List[int]): Indices of the objectives categorized to be
        decreased.
        lte_inds (List[int]): Indices of the objectives categorized to be
        reduced until some value is reached.
        rho (float): A small number to form the utopian point.
        rho_sum (float): A small number to be used as a weight for the sum
        term.

    .. _Miettinen 2006:
        Miettinen, K. & Mäkelä, Marko M.
        Synchronous approach in interactive multiobjective optimization
        European Journal of Operational Research, 2006, 170, 909-922

    """

    def __init__(
        self,
        nadir: np.ndarray,
        ideal: np.ndarray,
        lt_inds: List[int],
        lte_inds: List[int],
        rho: float = 1e-6,
        rho_sum: float = 1e-6,
    ):
        self.__nadir = nadir
        self.__ideal = ideal
        self.__lt_inds = lt_inds
        self.__lte_inds = lte_inds
        self.__rho = rho
        self.__rho_sum = rho_sum

    @property
    def nadir(self) -> np.ndarray:
        return self.__nadir

    @nadir.setter
    def nadir(self, val: np.ndarray):
        self.__nadir = val

    @property
    def ideal(self) -> np.ndarray:
        return self.__ideal

    @ideal.setter
    def ideal(self, val: np.ndarray):
        self.__ideal = val

    @property
    def lt_inds(self) -> List[int]:
        return self.__lt_inds

    @lt_inds.setter
    def lt_inds(self, val: List[int]):
        self.__lt_inds = val

    @property
    def lte_inds(self) -> List[int]:
        return self.__lte_inds

    @lte_inds.setter
    def lte_inds(self, val: List[int]):
        self.__lte_inds = val

    @property
    def rho(self) -> float:
        return self.__rho

    @rho.setter
    def rho(self, val: float):
        self.__rho = val

    @property
    def rho_sum(self) -> float:
        return self.__rho_sum

    @rho_sum.setter
    def rho_sum(self, val: float):
        self.__rho_sum = val

    def __call__(
        self, objective_vector: np.ndarray, reference_point: np.ndarray
    ) -> Union[float, np.ndarray]:
        # assure this function works with single objective vectors
        if objective_vector.ndim == 1:
            f = objective_vector.reshape((1, -1))
        else:
            f = objective_vector

        ii = self.lt_inds
        jj = self.lte_inds
        z = reference_point
        nad = self.nadir
        ide = self.ideal
        uto = self.ideal - self.rho

        lt_term = (f[:, ii] - ide[ii]) / (nad[ii] - uto[ii])
        lte_term = (f[:, jj] - z[jj]) / (nad[jj] - uto[jj])
        max_term = np.max(np.hstack((lt_term, lte_term)), axis=1)
        sum_term = self.rho_sum * np.sum(f / (nad - uto), axis=1)

        return max_term + sum_term


class StomASF(ASFBase):
    """Implementation of the satisficing trade-off method (STOM) as presented
    in `Miettinen 2006` equation (3.2)

    Args:
        ideal (np.ndarray): The ideal point.
        rho (float): A small number to form the utopian point.
        rho_sum (float): A small number to be used as a weight for the sum
        term.

    Attributes:
        ideal (np.ndarray): The ideal point.
        rho (float): A small number to form the utopian point.
        rho_sum (float): A small number to be used as a weight for the sum
        term.


    .. _Miettinen 2006:
        Miettinen, K. & Mäkelä, Marko M.
        Synchronous approach in interactive multiobjective optimization
        European Journal of Operational Research, 2006, 170, 909-922

    """

    def __init__(
        self, ideal: np.ndarray, rho: float = 1e-6, rho_sum: float = 1e-6
    ):
        self.__ideal = ideal
        self.__rho = rho
        self.__rho_sum = rho_sum

    @property
    def ideal(self) -> np.ndarray:
        return self.__ideal

    @ideal.setter
    def ideal(self, val: np.ndarray):
        self.__ideal = val

    @property
    def rho(self) -> float:
        return self.__rho

    @rho.setter
    def rho(self, val: float):
        self.__rho = val

    @property
    def rho_sum(self) -> float:
        return self.__rho_sum

    @rho_sum.setter
    def rho_sum(self, val: float):
        self.__rho_sum = val

    def __call__(
        self, objective_vectors: np.ndarray, reference_point: np.ndarray
    ) -> Union[float, np.ndarray]:
        # assure this function works with single objective vectors
        if objective_vectors.ndim == 1:
            f = objective_vectors.reshape((1, -1))
        else:
            f = objective_vectors

        z = reference_point
        uto = self.ideal - self.rho

        max_term = np.max((f - uto) / (z - uto), axis=1)
        sum_term = self.rho_sum * np.sum((f) / (z - uto), axis=1)

        return max_term + sum_term


class PointMethodASF(ASFBase):
    """Implementation of the reference point based ASF as presented
    in `Miettinen 2006` equation (3.3)

    Args:
        nadir (np.ndarray): The nadir point.
        ideal (np.ndarray): The ideal point.
        rho (float): A small number to form the utopian point.
        rho_sum (float): A small number to be used as a weight for the sum
        term.

    Note:
        Lack of better name...

    .. _Miettinen 2006:
        Miettinen, K. & Mäkelä, Marko M.
        Synchronous approach in interactive multiobjective optimization
        European Journal of Operational Research, 2006, 170, 909-922

    """

    def __init__(
        self,
        nadir: np.ndarray,
        ideal: np.ndarray,
        rho: float = 1e-6,
        rho_sum: float = 1e-6,
    ):
        self.__nadir = nadir
        self.__ideal = ideal
        self.__rho = rho
        self.__rho_sum = rho

    @property
    def nadir(self) -> np.ndarray:
        return self.__nadir

    @nadir.setter
    def nadir(self, val: np.ndarray):
        self.__nadir = val

    @property
    def ideal(self) -> np.ndarray:
        return self.__ideal

    @ideal.setter
    def ideal(self, val: np.ndarray):
        self.__ideal = val

    @property
    def rho(self) -> float:
        return self.__rho

    @rho.setter
    def rho(self, val: float):
        self.__rho = val

    @property
    def rho_sum(self) -> float:
        return self.__rho_sum

    @rho_sum.setter
    def rho_sum(self, val: float):
        self.__rho_sum = val

    def __call__(
        self, objective_vectors: np.ndarray, reference_point: np.ndarray
    ):
        # assure this function works with single objective vectors
        if objective_vectors.ndim == 1:
            f = objective_vectors.reshape((1, -1))
        else:
            f = objective_vectors

        z = reference_point
        nad = self.nadir
        uto = self.ideal - self.rho

        max_term = np.max((f - z) / (nad - uto), axis=1)
        sum_term = self.rho_sum * np.sum((f) / (nad - uto), axis=1)

        return max_term + sum_term


class AugmentedGuessASF(ASFBase):
    """Implementation of the augmented GUESS related ASF as presented in
    `Miettinen 2006` equation (3.4)

    Args:
        nadir (np.ndarray): The nadir point.
        ideal (np.ndarray): The ideal point.
        indx_to_exclude (List[int]): The indices of the objective functions to
        be excluded in calculating the first temr of the ASF.
        rho (float): A small number to form the utopian point.
        rho_sum (float): A small number to be used as a weight for the sum
        term.

    .. _Miettinen 2006:
        Miettinen, K. & Mäkelä, Marko M.
        Synchronous approach in interactive multiobjective optimization
        European Journal of Operational Research, 2006, 170, 909-922

    """

    def __init__(
        self,
        nadir: np.ndarray,
        ideal: np.ndarray,
        indx_to_exclude: List[int],
        rho: float = 1e-6,
        rho_sum: float = 1e-6,
    ):
        self.__nadir = nadir
        self.__ideal = ideal
        self.__indx_to_exclude = indx_to_exclude
        self.__rho = rho
        self.__rho_sum = rho_sum

    @property
    def nadir(self) -> np.ndarray:
        return self.__nadir

    @nadir.setter
    def nadir(self, val: np.ndarray):
        self.__nadir = val

    @property
    def ideal(self) -> np.ndarray:
        return self.__ideal

    @ideal.setter
    def ideal(self, val: np.ndarray):
        self.__ideal = val

    @property
    def indx_to_exclude(self) -> List[int]:
        return self.__indx_to_exclude

    @indx_to_exclude.setter
    def indx_to_exclude(self, val: List[int]):
        self.__indx_to_exclude = val

    @property
    def rho(self) -> float:
        return self.__rho

    @rho.setter
    def rho(self, val: float):
        self.__rho = val

    @property
    def rho_sum(self) -> float:
        return self.__rho_sum

    @rho_sum.setter
    def rho_sum(self, val: float):
        self.__rho_sum = val

    def __call__(
        self, objective_vectors: np.ndarray, reference_point: np.ndarray
    ):
        # assure this function works with single objective vectors
        if objective_vectors.ndim == 1:
            f = objective_vectors.reshape((1, -1))
        else:
            f = objective_vectors

        z = reference_point
        nad = self.nadir
        uto = self.ideal - self.rho
        ex_mask = np.full((f.shape[1]), True, dtype=bool)
        ex_mask[self.indx_to_exclude] = False

        max_term = np.max(
            (f[:, ex_mask] - nad[ex_mask]) / (nad[ex_mask] - z[ex_mask]),
            axis=1,
        )
        sum_term_1 = self.rho_sum * np.sum(
            (f[:, ex_mask]) / (nad[ex_mask] - z[ex_mask]), axis=1
        )
        # avoid division by zeros
        sum_term_2 = self.rho_sum * np.sum(
            (f[:, ~ex_mask]) / (nad[~ex_mask] - uto[~ex_mask]), axis=1
        )

        return max_term + sum_term_1 + sum_term_2



__all__ = [
    "ScalarConstraint",
    "ConstraintError",
    "constraint_function_factory",
]

from .Constraint import (
    ConstraintError,
    ScalarConstraint,
    constraint_function_factory,
)
